import abc
import logging
import warnings
from typing import Iterable, Union

import numpy as np
import tensorflow as tf

from uncertainty_wizard.internal_utils import UncertaintyWizardWarning
from uncertainty_wizard.models._stochastic._stochastic_mode import StochasticMode
from uncertainty_wizard.models._uwiz_model import _UwizModel
from uncertainty_wizard.models.stochastic_utils import layers
from uncertainty_wizard.models.stochastic_utils.layers import (
    UwizBernoulliDropout,
    UwizGaussianDropout,
    UwizGaussianNoise,
)
from uncertainty_wizard.quantifiers import Quantifier


class Stochastic(_UwizModel):
    """
    Stochastic models are models in which some randomness is added to the network during training.
    While this is typically done for network regularization, models trained in such a way can be used for
    uncertainty quantification. Simply speaking:

    Randomness (which is typically disabled during inference) can be enforced during inference,
    leading to predictions which are impacted by the random noise.
    By sampling multiple network outputs for the same input, we can infer the robustness of the network to the
    random noise. We assume that the higher the robustness, the higher the networks confidence.

    Instances of stochastic uncertainty wizard models can also be used in a non-stochastic way
    as as point prediction models (i.e., models without sampling)
    by calling the `model.predict` function
    or by passing a quantifier which does not rely on sampling to `model.predict_quantified` (such as Max-Softmax).
    Randomization during model inference is automatically enabled or disabled.
    """

    def __init__(self):
        """ABSTRACT METHOD. DO NOT CALL INIT DIRECTLY"""

    @property
    @abc.abstractmethod
    def inner(self) -> tf.keras.Model:
        """
        Direct access to the wrapped keras model.
        Use this if you want to directly work on the wrapped model.
        When using this, make sure not to modify the stochastic layers or the stochastic_mode tensor on the model.

        Returns: the tf.keras.Model wrapped in this StochasticSequential.
        """
        pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def stochastic_mode_tensor(self) -> tf.Variable:
        """
        Get access to the flag used to enable and disable the stochastic behavior.

        Returns: A boolean 0-dimensions tensorflow variable.

        """
        pass  # pragma: no cover

    def call(self, inputs, training=None, mask=None):
        """
        See tf.keras.Model.call for the documentation: The call is forwarded
        :param inputs: See tf.keras docs
        :param training: See tf.keras docs
        :param mask: See tf.keras docs
        :return: See tf.keras docs
        """
        return self.inner.call(inputs, training, mask)

    def compile(
        self,
        optimizer="rmsprop",
        loss=None,
        metrics=None,
        loss_weights=None,
        weighted_metrics=None,
        run_eagerly=None,
        expect_deterministic: bool = False,
    ):
        """
        This wraps the tf.keras.Model.compile method, but checks before if a stochastic layer was added to the model:
        If none was added, a warning is printed.

        This behavior can be turned of if you only intend to use the model as point predictor.
        In this case, set expect_deterministic to True.

        :param optimizer: See tf.keras.Model docs
        :param loss: See tf.keras.Model docs
        :param metrics: See tf.keras.Model docs
        :param loss_weights: See tf.keras.Model docs
        :param weighted_metrics: See tf.keras.Model docs
        :param run_eagerly: See tf.keras.Model docs
        :param expect_deterministic: Iff true, the model is not checked for randomness. Default: False
        :return: See tf.keras.Model docs
        """
        if not expect_deterministic:
            # Check is randomized
            is_stochastic = False
            for layer in self.inner.layers:
                if (
                    isinstance(layer, layers.UwizGaussianNoise)
                    or isinstance(layer, layers.UwizBernoulliDropout)
                    or isinstance(layer, layers.UwizGaussianDropout)
                ):
                    is_stochastic = True
                    break

            if not is_stochastic:
                warnings.warn(
                    "Looks like your model contains no uwiz random layers."
                    "If you added your own randomized layers, "
                    "or if you only want to use this with non-randomized point predictors, "
                    "you can savely ignore this warning."
                    "Otherwise, check the docs on uwiz stochastic models for more info.\n\n"
                    "Use compile(..., expect_deterministic=True) to disable this warning",
                    UncertaintyWizardWarning,
                )
        return self.inner.compile(
            optimizer=optimizer,
            loss=loss,
            metrics=metrics,
            loss_weights=loss_weights,
            weighted_metrics=weighted_metrics,
            run_eagerly=run_eagerly,
        )

    @property
    def fit(self):
        """
        Direct access to the fit method the wrapped keras model.
        See `tf.keras.Model.fit` for precise documentation of this method.

        Can be called as `stochastic_model.fit(...)`, equivalent to how fit would be called on a plain keras model.
        :return: The fit method of the wrapped model
        """
        return self.inner.fit

    @property
    def evaluate(self):
        """
        Direct access to the evaluate method the wrapped keras model.
        See `tf.keras.Model.evaluate` for precise documentation of this method.

        Can be called as `stochastic_model.evaluate(...)`, equivalent to how fit would be called on a plain keras model.
        This means that no stochastic sampling is done.

        :return: The evaluate method of the wrapped model
        """
        return self.inner.evaluate

    @property
    def predict(self):
        """
        Direct access to the predict method the wrapped keras model.
        See `tf.keras.Model.predict` for precise documentation of this method.

        Note that no confidences are calculated if calling this predict method, and the stochastic layers are disabled.
        To calculate confidences, call `model.predict_quantified(...)` instead of `model.predict(...)`

        Can be called as `model.predict(...)`,
        equivalent to how predict would be called on a plain keras model.
        :return: The predict method of the wrapped model
        """
        return self.inner.predict

    @property
    def summary(self):
        """
        Direct access to the summary method the wrapped keras model.
        See `tf.keras.Model.summary` for precise documentation of this method.
        """
        return self.inner.summary

    def save(
        self,
        filepath: str,
        overwrite: bool = True,
        include_optimizer: bool = True,
        save_format: str = None,
        signatures=None,
        options=None,
    ):
        """
        Save the model to file, as on plain tf models. Note that you must not use the h5 file format.

        ** Attention ** uwiz models must be loaded using `uwiz.models.load_model` AND NOT using the corresponding
        keras method.

        See below the keras documentation, with applies for this method as well
        (taking in account the limitations mentioned above)

        """
        # Append the keras documentation
        Stochastic.save.__doc__ += tf.keras.Model.save.__doc__
        assert (
            not filepath.lower().endswith("h5")
            and not filepath.lower().endswith("hdf5")
            and not filepath.lower().endswith(".keras")
            and (save_format is None or not save_format.lower() == "h5")
        ), (
            "Uncertainty Wizard does not support the deprecated h5 format to save models."
            "Change the file ending or the save_format parameter to save using the better tf `SavedModel` format."
        )

        return self.inner.save(
            filepath=filepath,
            overwrite=overwrite,
            include_optimizer=include_optimizer,
            save_format=save_format,
            signatures=signatures,
            options=options,
        )

    def _get_stochastic_mode(self):
        """
        Get access to the stochastic mode used in the model to toggle randomness during predictions.
        :return: A stochastic mode instance wrapping the stochastic mode tensor used in this model.
        """
        return StochasticMode(self.stochastic_mode_tensor)

    @classmethod
    def _replace_layer_if_possible(
        cls, layer, stochastic_mode
    ) -> tf.keras.layers.Layer:
        if isinstance(layer, tf.keras.layers.Dropout):
            return UwizBernoulliDropout.from_keras_layer(
                layer=layer, stochastic_mode=stochastic_mode
            )
        elif isinstance(layer, tf.keras.layers.GaussianNoise):
            return UwizGaussianNoise.from_keras_layer(
                layer=layer, stochastic_mode=stochastic_mode
            )
        if isinstance(layer, tf.keras.layers.GaussianDropout):
            return UwizGaussianDropout.from_keras_layer(
                layer=layer, stochastic_mode=stochastic_mode
            )
        else:
            # The passed layer is not replaceable with a stochastic layer
            return layer

    def _get_scores(
        self, x, batch_size: int, verbose, steps, sample_size
    ) -> np.ndarray:
        if isinstance(x, tf.data.Dataset):
            logging.debug(
                "You passed a tf.data.Dataset to predict_quantified in a stochastic model."
                "tf.data.Datasets passed to this method must not be batched. We take care of the batching."
                "Please make sure that your dataset is not batched (we can not check that)"
            )
            x_as_ds = x
        elif isinstance(x, np.ndarray):
            x_as_ds = tf.data.Dataset.from_tensor_slices(x)
        else:
            raise ValueError(
                "At the moment, uwiz stochastic models support only (unbatched)"
                "numpy arrays and tf.data.Datasets as inputs. "
                "Please transform your input in one of these forms."
            )

        # Repeat every input `sample_size` many times in-place
        num_samples_tensor = tf.reshape(tf.constant(sample_size), [1])

        @tf.function
        @tf.autograph.experimental.do_not_convert
        def _expand_to_sample_size(inp):
            shape = tf.concat((num_samples_tensor, tf.shape(inp)), axis=0)
            return tf.broadcast_to(input=inp, shape=shape)

        inputs = x_as_ds.map(_expand_to_sample_size).unbatch()

        # Batch the resulting dataset
        inputs = inputs.batch(batch_size=batch_size)

        # Make predictions.
        if steps is not None:
            steps = steps * sample_size
        outputs = self.inner.predict(inputs, verbose=verbose, steps=steps)

        # Reshape sampled predictions, grouping by sample (such that outputs[i] contains all samples for input i
        output_shape = list(outputs.shape)
        output_shape.insert(0, -1)
        output_shape[1] = sample_size
        return outputs.reshape(output_shape)

    def predict_quantified(
        self,
        x: Union[tf.data.Dataset, np.ndarray],
        quantifier: Union[Quantifier, str, Iterable[Union[str, Quantifier]]],
        # Other Sequential.predict params (e.g. Callbacks) are not yet supported
        sample_size: int = 64,
        batch_size: int = 32,
        verbose: int = 0,
        steps=None,
        as_confidence: Union[None, bool] = None,
    ):
        """
        Calculates predictions and uncertainties (or confidences) according to the passed quantifer(s).
        Sampling is done internally.
        Both point-predictor and sampling based quantifiers can be used in the same method call.
        Uwiz automatically enables and disables the randomness of the model accordingly.
        :param x: The inputs for which the predictions should be made. tf.data.Dataset (unbatched) or numpy array.
        :param quantifier: The quantifier or quantifier alias to use (or a collection of them)
        :param sample_size: The number of samples to be used for sample-expecting quantifiers
        :param batch_size: The batch size to be used for predictions
        :param verbose: Prediction process logging, as in tf.keras.Model.fit
        :param steps: Predictions steps, as in tf.keras.Model.fit. Is adapted according to chosen sample size.
        :param as_confidence: If true, uncertainties are multiplied by (-1),
        if false, confidences are multiplied by (-1). Default: No transformations.
        :return: A tuple (predictions, uncertainties_or_confidences) if a single quantifier was
        passed as string or instance, or a collection of such tuples if the passed quantifiers was an iterable.
        """
        all_q, pp_q, sample_q, return_single_tuple = self._quantifiers_as_list(
            quantifier
        )
        self._check_quantifier_heterogenity(
            as_confidence=as_confidence, quantifiers=all_q
        )

        self._warn_if_invalid_sample_size(
            sample_size, samples_based_quantifiers=sample_q
        )

        stochastic_scores, point_prediction_scores = None, None
        if len(sample_q) > 0:
            self.stochastic_mode_tensor.assign(True)
            stochastic_scores = self._get_scores(
                x, batch_size, verbose, steps, sample_size
            )
            self.stochastic_mode_tensor.assign(False)
        if len(pp_q) > 0:
            if isinstance(x, tf.data.Dataset):
                x = x.batch(batch_size=batch_size)
            point_prediction_scores = self.predict(
                x, batch_size=batch_size, verbose=verbose, steps=steps
            )

        results = self._run_quantifiers(
            as_confidence, point_prediction_scores, all_q, stochastic_scores
        )
        if return_single_tuple:
            return results[0]
        return results

    @staticmethod
    def _run_quantifiers(
        as_confidence, point_prediction_scores, quantifiers, stochastic_scores
    ):
        results = []
        for q in quantifiers:
            if q.takes_samples():
                assert stochastic_scores is not None, (
                    "Uncertainty Wizard internal error. "
                    "Did not compute stochastic scores"
                )
                scores = stochastic_scores
            else:
                assert point_prediction_scores is not None, (
                    "Uncertainty Wizard internal error. "
                    "Did not compute point prediction scores"
                )
                scores = point_prediction_scores
            predictions, superv_scores = q.calculate(scores)
            superv_scores = q.cast_conf_or_unc(
                as_confidence=as_confidence, superv_scores=superv_scores
            )
            results.append((predictions, superv_scores))
        return results

    @staticmethod
    def _warn_if_invalid_sample_size(sample_size, samples_based_quantifiers):
        if len(samples_based_quantifiers) > 0 and sample_size < 2:
            warnings.warn(
                f"The sample_size parameter must be greater than 1, but was {sample_size}",
                category=UncertaintyWizardWarning,
                # This shows the warning in the users code where he does the wrong call
                stacklevel=3,
            )

    @classmethod
    @abc.abstractmethod
    def _wrap(cls, inner, stochastic_mode_tensor=None):
        """
        Method to wrap a passed model.
        The model must already rely on a stochastic mode tensor which either has to be passed as a param
        or has to be already attached to the model as attribute.

        Args:
            inner (): The model to wrap
            stochastic_mode_tensor (): The stochastic mode to attach on the wrapped model, if not yet done

        Returns:
            An instance of the implementing subclass (i.e., a subclass of _AbstractStochastic)
        """
        # pragma: no cover



import warnings
from typing import Iterable, Union

from uncertainty_wizard.internal_utils import UncertaintyWizardWarning
from uncertainty_wizard.quantifiers import Quantifier, QuantifierRegistry


class _UwizModel:
    @staticmethod
    def _quantifiers_as_list(quantifier):
        is_single_quantifier = False
        if isinstance(quantifier, list):
            quantifiers = quantifier
        else:
            quantifiers = [quantifier]
            is_single_quantifier = True

        quantifier_objects = []
        for quantifier in quantifiers:
            if isinstance(quantifier, str):
                quantifier_objects.append(QuantifierRegistry.find(quantifier))
            elif isinstance(quantifier, Quantifier):
                quantifier_objects.append(quantifier)
            else:
                raise TypeError(
                    "The passed quantifier is neither a quantifier instance nor a quantifier alias (str)."
                    f"Type of the passed object {str(type(quantifier))}"
                )

        point_prediction_quantifiers = [
            q for q in quantifier_objects if q.takes_samples() is False
        ]
        samples_based_quantifiers = [
            q for q in quantifier_objects if q.takes_samples() is True
        ]
        return (
            quantifier_objects,
            point_prediction_quantifiers,
            samples_based_quantifiers,
            is_single_quantifier,
        )

    @staticmethod
    def _check_quantifier_heterogenity(
        as_confidence: Union[None, bool], quantifiers: Iterable[Quantifier]
    ) -> None:
        if as_confidence is None:
            num_uncertainties = sum(q.is_confidence() is False for q in quantifiers)
            num_confidences = sum(q.is_confidence() is True for q in quantifiers)
            if num_confidences > 0 and num_uncertainties > 0:
                warnings.warn(
                    """ 
                    You are predicting both confidences and uncertainties.
                    When comparing the two, keep in mind that confidence is expected to 
                    correlate positively with the probability of a correct prediction,
                    while uncertainty is expected to correlate negatively.
    
                    You may want to use the `as_confidence` flag when calling `predict_quantified`: 
                    If set to false, it multiplies confidences by (-1). 
                    If set to true, it multiplies uncertainties by (-1).
    
                    See uwiz.Quantifier.cast_conf_or_unc for more details.
                    """,
                    category=UncertaintyWizardWarning,
                    stacklevel=3,
                )



from typing import List, Tuple

import numpy as np

from uncertainty_wizard.quantifiers.quantifier import ProblemType, UncertaintyQuantifier


class VariationRatio(UncertaintyQuantifier):
    """
    A predictor & uncertainty quantifier, based on multiple samples (e.g. nn outputs) in a classification problem

    The prediction is made using a plurality vote, i.e., the class with the highest value in most samples is selected.
    In the case of a tie, the class with the lowest index is selected.

    The uncertainty is quantified using the variation ratio `1 - w / S`,
    where w is the number of samples where the overall prediction equals the prediction of the sample
    and S is the total number of samples.
    """

    # docstr-coverage:inherited
    @classmethod
    def takes_samples(cls) -> bool:
        return True

    # docstr-coverage:inherited
    @classmethod
    def problem_type(cls) -> ProblemType:
        return ProblemType.CLASSIFICATION

    # docstr-coverage:inherited
    @classmethod
    def aliases(cls) -> List[str]:
        return ["variation_ratio", "vr", "var_ratio"]

    # docstr-coverage:inherited
    @classmethod
    def calculate(cls, nn_outputs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        assert len(nn_outputs.shape) == 3, (
            "nn_outputs for this quantifier must have shape "
            "(num_inputs, num_samples, num_classes)"
        )
        num_inputs = nn_outputs.shape[0]
        num_samples = nn_outputs.shape[1]
        num_classes = nn_outputs.shape[2]

        sofmax_table = np.reshape(nn_outputs, (num_inputs * num_samples, num_classes))
        per_sample_argmax = np.argmax(sofmax_table, axis=1)

        is_max_array = np.zeros(
            shape=(num_inputs * num_samples, num_classes), dtype=bool
        )
        is_max_array[np.arange(num_inputs * num_samples), per_sample_argmax] = True

        per_input_is_max_array = np.reshape(
            is_max_array, (num_inputs, num_samples, num_classes)
        )

        sum_array_dtype = cls._sum_array_dtype(num_samples)
        votes_counts = np.sum(per_input_is_max_array, axis=1, dtype=sum_array_dtype)

        predictions = cls._prediction_array_with_appropriate_dtype(
            num_classes, num_inputs
        )
        np.argmax(votes_counts, axis=1, out=predictions)
        max_counts = votes_counts[np.arange(num_inputs), predictions]

        vr = 1 - max_counts / num_samples
        return predictions, vr

    @classmethod
    def _sum_array_dtype(cls, num_samples):
        """Selects an appropriate dtype (np.uint16 or np.uint8) based on the number of samples """
        # uint16 allows up to 65535 samples (which is way above reasonable for any problem)
        # If there are up to 255 samples per input, we can save half the memory using uint8
        sum_array_dtype = np.uint16
        if num_samples < 256:
            sum_array_dtype = np.uint8
        return sum_array_dtype

    @classmethod
    def _prediction_array_with_appropriate_dtype(cls, num_classes, num_inputs):
        """Creates an empty one-dimensional array with the number of inputs as length
        and an appropriate dtype (np.uint16 or np.uint8).
        This can then be used to store the predictions"""
        # uint16 allows up to 65535 samples (which is way above reasonable for any problem)
        # If there are up to 255 samples per input, we can save half the memory using uint8
        predictions = np.empty(shape=num_inputs, dtype=np.uint16)
        if num_classes < 256:
            predictions = np.empty(shape=num_inputs, dtype=np.uint8)
        return predictions



import gc
from dataclasses import dataclass
from typing import Dict, Tuple, Union

import numpy as np
import tensorflow as tf


@dataclass
class DataLoadedPredictor:
    """
    The default task to be executed for predictions where the input data is a numpy array.
    Leaves the serialization and deserialization of the array to the python multiprocessing library,
    and does thus not explicitly implement it here.
    """

    x_test: np.ndarray
    batch_size: int
    steps: int = None

    def __call__(self, model_id: int, model: tf.keras.Model):
        """Simple call to keras predict, formulated as __call__ to allow for constructor params."""
        return model.predict(
            x=self.x_test, batch_size=self.batch_size, steps=self.steps, verbose=1
        )


@dataclass
class NumpyFitProcess:
    """
    This is a class used as callable for the serialization and deserialization of numpy arrays
    which are then used in the keras fit process.
    """

    x: Union[str, np.ndarray] = None
    y: Union[str, np.ndarray] = None
    batch_size: int = None
    epochs: int = 1
    verbose: int = 1
    # Callbacks not supported in this default process (as type does not guarantee picklability)
    # callbacks = None,
    validation_split: float = 0.0
    validation_data: Union[Tuple[str, str], Tuple[np.ndarray, np.ndarray]] = None
    shuffle: bool = True
    class_weight: Dict[int, float] = None
    sample_weight: np.ndarray = None
    initial_epoch: int = 0
    steps_per_epoch: int = None
    validation_steps: int = None
    validation_freq: int = 1

    # Max_queue_size, workers and use_multiprocessing not supported as we force input to be numpy array
    # max_queue_size = 10,
    # workers = 1,
    # use_multiprocessing = False

    def __call__(
        self, model_id: int, model: tf.keras.Model
    ) -> Tuple[tf.keras.Model, tf.keras.callbacks.History]:
        """Simple call to keras fit, formulated as __call__ to allow for constructor params."""
        x = np.load(self.x, allow_pickle=True) if isinstance(self.x, str) else self.x
        y = np.load(self.y, allow_pickle=True) if isinstance(self.y, str) else self.y
        if self.validation_data is not None and isinstance(
            self.validation_data[0], str
        ):
            val_x = np.load(self.validation_data[0], allow_pickle=True)
            val_y = np.load(self.validation_data[1], allow_pickle=True)
            val_data = (val_x, val_y)
        else:
            val_data = self.validation_data
        history = model.fit(
            x=x,
            y=y,
            batch_size=self.batch_size,
            epochs=self.epochs,
            verbose=self.verbose,
            validation_split=self.validation_split,
            validation_data=val_data,
            shuffle=self.shuffle,
            class_weight=self.class_weight,
            sample_weight=self.sample_weight,
            initial_epoch=self.initial_epoch,
            steps_per_epoch=self.steps_per_epoch,
            validation_steps=self.validation_steps,
            validation_freq=self.validation_freq,
        )
        del x
        del y
        if val_data:
            del val_data
        gc.collect()
        return model, history.history



from typing import List

import numpy as np

from .quantifier import ProblemType, UncertaintyQuantifier


def validate_shape(nn_outputs):
    """Makes sure the nn outputs is a valid input for the regression classifiers in this file."""
    assert len(nn_outputs.shape) >= 2, (
        "nn_outputs for Average Standard Deviation must have shape "
        "(number_of_inputs, num_samples, pred_dim_1, pred_dim_2, ...)"
    )
    num_inputs = nn_outputs.shape[0]
    num_samples = nn_outputs.shape[1]
    return num_inputs, num_samples


class StandardDeviation(UncertaintyQuantifier):
    """
    Measures the standard deviation over different samples of a regression problem, i.e., an arbitrary problem,
    which is used as Uncertainty and the mean of the samples as prediction

    This implementation can handle both regression prediction consisting of a single scalar dnn output
    as well as larger-shaped dnn outputs. In the latter case, entropy is calculated and returned
    for every position in the dnn output shape.
    """

    # docstr-coverage:inherited
    @classmethod
    def takes_samples(cls) -> bool:
        return True

    # docstr-coverage:inherited
    @classmethod
    def problem_type(cls) -> ProblemType:
        return ProblemType.REGRESSION

    # docstr-coverage:inherited
    @classmethod
    def aliases(cls) -> List[str]:
        return ["standard_deviation", "std_dev", "std", "stddev", "StandardDeviation"]

    # docstr-coverage:inherited
    @classmethod
    def calculate(cls, nn_outputs: np.ndarray):
        _, _ = validate_shape(nn_outputs)
        predictions = np.mean(nn_outputs, axis=1)
        uncertainties = np.std(nn_outputs, axis=1)
        return predictions, uncertainties



import logging
import os
import shutil


def _preprocess_path(path: str) -> str:
    """
    Preprocesses the path, i.e.,
    it makes sure there's no tailing '/' or '\'
    :param path: The path as passed by the user
    :return: The path of the folder on which the model will be stored
    """
    if path.endswith("/") or path.endswith("\\"):
        path = path[: len(path) - 1]
    return path


class SaveConfig:
    """
    This is a class containing information and utility methods about the saving of the ensemble.
    Currently, it only contains a few static fields. This may change in the future.

    Instances of the SaveConfig can be used in the save_single_model and load_single_model
     methods of EnsembleContextManagers. However, consider SaveConfigs as read-only classes.
     In addition, instances should only be created internally by uncertainty wizard
     and not in custom user code.

    """

    def __init__(
        self, ensemble_save_base_path: str, delete_existing: bool, expect_model: bool
    ):
        """
        ** WARNING ** Typically, there's no need for uwiz users to call this contructor:
        They are created by uncertainty wizard and passed to (potentially custom made) context managers.

        Creates a new save config.
        Note that this automatically triggers the preparation of the specified save_path
        (e.g. the deletion of existing
        :param ensemble_save_base_path: Where to store the ensemble
        :param delete_existing: If the folder is non empty and delete_existing is True,
         the folder will be cleared. If this is false and the folder is non-empty,
         a warning will be logged.
        :param expect_model: If this is True, we expect that there is already an
         ensemble file in the folder and the warning described for 'delete_existing'
         will not be logged.
        """
        self._ensemble_save_base_path = _preprocess_path(ensemble_save_base_path)
        self._create_or_empty_folder(
            path=ensemble_save_base_path,
            overwrite=delete_existing,
            expect_model=expect_model,
        )

    def filepath(self, model_id: int):
        """
        This methods builds the path on which a particular atomic model should be saved / found
        :param model_id: the id of the atomic model
        :return: A path, where a folder named after the model id is appended to self.ensemble_save_base_path
        """
        return f"{self._ensemble_save_base_path}/{model_id}"

    @property
    def ensemble_save_base_path(self) -> str:
        """
        Returns the path (as string) where this ensemble is stored.
        This path is always a folder and after successful creation of the ensemble,
        it will contain the ensemble config file and a subfolder for every atomic model.
        :return: The save path of this ensemble as string
        """
        return self._ensemble_save_base_path

    @staticmethod
    def _create_or_empty_folder(path: str, overwrite: bool, expect_model=False) -> None:
        if os.path.exists(path=path):
            if overwrite:
                shutil.rmtree(path, ignore_errors=False, onerror=None)
                os.mkdir(path)
            elif not expect_model:
                logging.info(
                    f"A folder {path} already exists. "
                    f"We will assume it contains a lazy ensemble. "
                    f"Specify `delete_existing=True` to empty the folder when creating the LazyEnsemble."
                )
        else:
            os.makedirs(path)



from typing import List, Tuple

import numpy as np

from uncertainty_wizard.quantifiers.one_shot_classifiers import MaxSoftmax
from uncertainty_wizard.quantifiers.quantifier import ConfidenceQuantifier, ProblemType


class MeanSoftmax(ConfidenceQuantifier):
    """
    A predictor & uncertainty quantifier, based on multiple samples (e.g. nn outputs) in a classification problem.

    Both the prediction and the uncertainty score are calculated using the average softmax values over all samples.
    This is sometimes also called 'ensembling', as it is often used in deep ensembles.
    """

    # docstr-coverage:inherited
    @classmethod
    def aliases(cls) -> List[str]:
        return ["mean_softmax", "ensembling", "ms"]

    # docstr-coverage:inherited
    @classmethod
    def takes_samples(cls) -> bool:
        return True

    # docstr-coverage:inherited
    @classmethod
    def problem_type(cls) -> ProblemType:
        return ProblemType.CLASSIFICATION

    # docstr-coverage:inherited
    @classmethod
    def calculate(cls, nn_outputs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:

        # For simplicity, we let the predictions be calculated by the Variation Ratio Code
        #    accepting a slight overhead from also calculating the actual variation ratio
        assert len(nn_outputs.shape) == 3, (
            "nn_outputs for this quantifier must have shape "
            "(num_inputs, num_samples, num_classes)"
        )

        # Take means over the samples
        means = np.mean(nn_outputs, axis=1)

        # Calculate argmax as prediction and max as confidence
        return MaxSoftmax.calculate(means)



"""
Uncertainty wizard models and corresponding utilities
"""

__all__ = [
    # Models
    "Stochastic",
    "StochasticSequential",
    "StochasticFunctional",
    "LazyEnsemble",
    # Helper Objects
    "StochasticMode",
    # Model factories
    "stochastic_from_keras",
    "load_model",
]
from ._load_model import load_model
from ._stochastic._abstract_stochastic import Stochastic
from ._stochastic._from_keras import stochastic_from_keras
from ._stochastic._functional_stochastic import StochasticFunctional
from ._stochastic._sequential_stochastic import StochasticSequential
from ._stochastic._stochastic_mode import StochasticMode
from .ensemble_utils._lazy_ensemble import LazyEnsemble



import tensorflow as tf
from tensorflow.python.platform import tf_logging as logging

from uncertainty_wizard.models._stochastic._abstract_stochastic import Stochastic
from uncertainty_wizard.models._stochastic._stochastic_mode import StochasticMode


class StochasticSequential(Stochastic):
    """
    A stochastic wrapper of `tf.keras.models.Sequential` models, suitable for MC Dropout
    and similar sampling based approaches on randomized models.
    """

    # Include the superclass documentation
    __doc__ += Stochastic.__doc__

    # docstr-coverage:inherited
    def __init__(self, layers=None, name=None):
        super().__init__()
        # Create an empty sequential model with a stochastic mode
        self._inner_sequential = tf.keras.models.Sequential(name=name)
        self._inner_sequential._stochastic_mode_tensor = StochasticMode().as_tensor()
        # Append all the passed layers
        if layers is not None:
            for layer in layers:
                self.add(layer)

    @classmethod
    def _wrap(cls, inner, stochastic_mode_tensor=None):
        model = StochasticSequential()
        model._inner_sequential = inner
        if stochastic_mode_tensor is None:
            assert model._inner_sequential._stochastic_mode_tensor is not None, (
                "Uncertainty Wizard internal error. "
                "Trying to wrap a model that has no stochastic_mode_tensor, "
                "and no external stochastic_mode_tensor is passed to attach"
            )
        else:
            model._inner_sequential._stochastic_mode_tensor = stochastic_mode_tensor
        return model

    # docstr-coverage:inherited
    @property
    def inner(self):
        return self._inner_sequential

    # docstr-coverage:inherited
    @property
    def stochastic_mode_tensor(self):
        return self._inner_sequential._stochastic_mode_tensor

    def add(self, layer, prevent_use_for_sampling=False):
        """
        Adds the layer to the model. See docs of `tf.keras.model.Sequential.add(layer)` for details.

        In addition, layers of type
        `tf.keras.layers.Dropout`,
        `tf.keras.layers.GaussianNoise` and
        `tf.keras.layers.GaussianDropout`
        are overridden by equivalent layers which allow to be enabled during inference for randomized predictions.

        Arguments:
            layer: layer instance to be added to the model.
            prevent_use_for_sampling: Do not use the layer for randomization during inference. Has only effect on layers of type `Dropout`, `GaussianNoise` or `GaussianDropout`
        """
        # Add noise layer, if applicable
        if not prevent_use_for_sampling:
            layer = self._replace_layer_if_possible(
                layer, stochastic_mode=self._get_stochastic_mode()
            )
        # Add normal as defined by user
        self.inner.add(layer)

    def get_config(self):
        """
        Not supported
        :return: An empty config
        """
        logging.warning(
            """ 
            It looks like you are trying to serialize a StochasticSequential model.
            Please note that to save an StochasticSequential model, you have to call `model.save(...)`
            and to load it, you have to use `StochasticSequential.load_model(...)`
            """
        )
        return []



import json
import multiprocessing
import os
import tempfile
import warnings
from functools import partial
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, TypeVar, Union

import numpy as np
import tensorflow as tf

from uncertainty_wizard.models._uwiz_model import _UwizModel
from uncertainty_wizard.models.ensemble_utils import (
    DynamicGpuGrowthContextManager,
    EnsembleContextManager,
    NoneContextManager,
    SaveConfig,
)
from uncertainty_wizard.models.ensemble_utils._callables import (
    DataLoadedPredictor,
    NumpyFitProcess,
)
from uncertainty_wizard.models.ensemble_utils._save_config import _preprocess_path
from uncertainty_wizard.quantifiers import Quantifier

T = TypeVar("T")


def _model_creation_and_saving_process(
    model_id: int,
    save_config: SaveConfig,
    inner_func: Callable[[int], Tuple[tf.keras.Model, T]],
    context: Callable[[int, dict], EnsembleContextManager],
):
    # No varargs yet. This is a placeholder for future version extensions
    varargs = dict()
    # No custom methods called on ctxt yet. But this may change in future versions.
    with context(model_id, varargs) as ctxt:
        model, ret_container = inner_func(model_id)
        ctxt.save_single_model(model_id=model_id, model=model, save_config=save_config)
    return ret_container


def _model_updating_process(
    model_id: int,
    save_config: SaveConfig,
    inner_func: Callable[[int, tf.keras.Model], Tuple[tf.keras.Model, Any]],
    context: Callable[[int, dict], EnsembleContextManager],
):
    # No varargs yet. This is a placeholder for future version extensions
    varargs = dict()
    # No custom methods called on ctxt yet. But this may change in future versions.
    with context(model_id, varargs) as ctxt:
        model = ctxt.load_single_model(model_id, save_config=save_config)
        model, ret_container = inner_func(model_id, model)
        ctxt.save_single_model(model_id=model_id, model=model, save_config=save_config)
    return ret_container


def _model_consuming_process(
    model_id: int,
    save_config: SaveConfig,
    inner_func: Callable[[int, tf.keras.Model], Any],
    context: Callable[[int, dict], EnsembleContextManager],
):
    # No varargs yet. This is a placeholder for future version extensions
    varargs = dict()
    # No custom methods called on ctxt yet. But this may change in future versions.
    with context(model_id, varargs) as ctxt:
        model = ctxt.load_single_model(model_id, save_config=save_config)
        ret_container = inner_func(model_id, model)
    return ret_container


def config_file_path(path):
    """
    Constructs the path of the config file for ensemble saving.
    Args:
        path (): The path of the folder in which the ensemble is going to be saved or was saved.
    Returns: The path of the save_config file.
    """
    # This will be a repeated preprocessing in most cases,
    #   but we do it every time nonetheless, to avoid nasty bugs due to forgotten preprocessing

    path = _preprocess_path(path)

    return f"{path}/lazy_ensemble_config.pickle"


def _store_to_temp_file(x: np.ndarray):
    with tempfile.NamedTemporaryFile(delete=False, suffix=".npy") as file:
        file_name = file.name
    np.save(file_name, x)
    return file_name


class LazyEnsemble(_UwizModel):
    """
    LazyEnsembles are uncertainty wizards implementation of Deep Ensembles,
    where multiple atomic models are trained for the same problem;
    the output distribution (and thus uncertainty) is then inferred from predicting on all atomic models.

    ** Multi-Processing **

    This ensemble implementation is lazy as it does not keep the atomic models in memory
    (or even worse, in the tf graph).
    Instead, atomic models are persisted on the file system and only loaded when needed -
    and discarded immediately afterwards.
    To further increase performance, in particular on high performance GPU powered hardware setups,
    where a single model instance training does not use the full GPU resources,
    LazyEnsemble allows to create multiple concurrent tensorflow sessions, each running a dedicated model in parallel.
    The number of processes to be used can be specified on essentially any LazyEnsemble function.


    Models are loaded into a context, e.g. a gpu configuration which was configured before the model was loaded.
    The default context, if multiple processes are used, sets the GPU usage to dynamic memory growth.
    Pay attention: By using too many processes, it is easy to exhaust your systems resources.
    Thus, we recommend to set the number of processes conservatively, observe the system load
    and increase the number of processes if possible.
    Default contexts are uwiz.models.ensemble_utils.DynamicGpuGrowthContextManager if multiprocessing is enables,
    and uwiz.models.ensemble_utils.NoneContextManager otherwise.

    Note: Multi-Processing can be disabled by setting the number of processes to 0.
    Then, predictions will be made in the main process on the main tensorflow session.
    *Attention*: In this case, the tensorflow session will be cleared after every model execution!

    ** The LazyEnsemble Interface & Workflow **

    LazyEnsemble exposes four central functions: create, modify and consume.
    In general, every of these functions expectes a picklable function at input
    which either creates, modifies or consumes a plain keras model.
    Please refer to the specific methods documentation for details.
    Furthermore LazyEnsemble exposes utility methods wrapping the above listed methods,
    e.g. fit and predict_quantified, which expect numpy array inputs and automatically
    serialize and deserialize them to be used in parallel processes.

    ** Stability of Lazy Ensembles **

    To optimize GPU use, LazyEnsemble relies on some of tensorflows experimental features and is thus,
    by extension, also to be considered experimental.

    """

    def __init__(
        self,
        num_models: int,
        model_save_path: str,
        delete_existing: bool = True,
        expect_model: bool = False,
        default_num_processes: int = 1,
    ):
        """
        Creates a new lazy model.
        There's not much done in this method, actually.
        Essentially, this opens and prepares the folder in which the atomic models will be stored.
        :param num_models: The number of models to be used.
        :param model_save_path: The path (folder) where the models should be stored.
        :param delete_existing: If True (default) the content of the specified folder will be cleared.
        :param expect_model: If false, delete existing is false and the specified folder is not empty,
            a warning will be printed.
        :param default_num_processes: The default number of processes to use. This can be overridden in later calls.
        """
        self.default_num_processes = default_num_processes
        self.num_models = num_models
        self.save_config = SaveConfig(
            ensemble_save_base_path=model_save_path,
            expect_model=expect_model,
            delete_existing=delete_existing,
        )
        with open(config_file_path(path=model_save_path), "w") as f:
            json.dump({"num_models": num_models}, f)

    def _num_processes_or_default(self, num_processes: Optional[int]):
        return self.default_num_processes if num_processes is None else num_processes

    def create(
        self,
        create_function: Callable[[int], Tuple[tf.keras.Model, T]],
        num_processes: int = None,
        context: Callable[[int, dict], EnsembleContextManager] = None,
    ) -> List[T]:
        """
        This function takes care of the creation of new atomic models for this ensemble instance.
        At its core stands a create_function:
        This custom function takes as input the id of the model to be generated (which may be ignored),
        and is expected to return the newly created keras model and some custom,
        picklable, creation report (e.g. the fit history).
        If not required, the returned report may be None.
        You should refrain from returning extremely large report object,
        as they will be kept in memory and may occupy too many system resources.

        :param create_function: A picklable function to create new atomic models,
            as explained in the description above.
        :param num_processes: The number of processes to use.
            Default: The default or value specified when creating the lazy ensemble.
        :param context: A contextmanager which prepares a newly crated process for execution
            (e.g. by configuring the gpus). See class docstring for explanation of default values.
        :return: The reports returned by the create_function executions.
        """
        return self._run_in_processes(
            process_creator=_model_creation_and_saving_process,
            inner_function=create_function,
            num_processes=num_processes,
            context=context,
        )

    def modify(
        self,
        map_function: Callable[[int, tf.keras.Model], Tuple[tf.keras.Model, T]],
        num_processes: int = None,
        context: Callable[[int], EnsembleContextManager] = None,
    ) -> List[T]:
        """
        This function takes care of modifications of previously generated atomic models for this ensemble instance.
        At its core stands a map_function:
        This custom function takes as input the id of the model to be modified (which may be ignored)
        and the model instance.
        It is expected to return the modified (or replaced) keras model and some custom,
        picklable, modification report (e.g. the fit history).
        If not required, the returned report may be None.
        You should refrain from returning extremely large report object,
        as they will be kept in memory and may occupy too many system resources.

        *Attention* Whenever possible, try to reduce the number of calls to this function.
        For example, it is often possible to train models as part of the 'create' call.
        This will result in the creation of less processes and thus a faster overall performance.

        :param map_function: A picklable function to modify atomic models,
            as explained in the description above.
        :param num_processes: The number of processes to use.
            Default: The default or value specified when creating the lazy ensemble.
        :param context: A contextmanager which prepares a newly crated process for execution
            (e.g. by configuring the gpus). See class docstring for explanation of default values.
        :return: The reports returned by the create_function executions.
        """
        return self._run_in_processes(
            process_creator=_model_updating_process,
            inner_function=map_function,
            num_processes=num_processes,
            context=context,
        )

    def consume(
        self,
        consume_function: Callable[[int, tf.keras.Model], T],
        num_processes: int = None,
        context: Callable[[int], EnsembleContextManager] = None,
    ) -> List[T]:
        """
        This function uses the atomic models in the ensemble without changing them.
        At its core stands a consume_function:
        This custom function takes as input the id of the model to be modified (which may be ignored)
        and the model instance.
        It is expected to return a picklable consumption result.
        You should refrain from returning extremely large consumption results,
        as they will be kept in memory and may occupy too many system resources.
        In such a case, you may want to persist the results and return None as a consumption result instead.

        *Attention* While this function can be used for predictions, you'd probably prefer to use
        ensemble.quantify_predictions(...) instead, which wraps this functions and allows to apply
        quantifiers for overall prediction inference and uncertainty quantification.

        :param consume_function: A picklable function to consume atomic models,
            as explained in the description above.
        :param num_processes: The number of processes to use.
            Default: The default or value specified when creating the lazy ensemble.
        :param context: A contextmanager which prepares a newly crated process for execution
            (e.g. by configuring the gpus). See class docstring for explanation of default values.
        :return: The reports returned by the create_function executions.
        """
        return self._run_in_processes(
            process_creator=_model_consuming_process,
            inner_function=consume_function,
            num_processes=num_processes,
            context=context,
        )

    def fit(
        self,
        x: np.ndarray = None,
        y: np.ndarray = None,
        batch_size: int = None,
        epochs: int = 1,
        verbose: int = 1,
        callbacks=None,
        validation_split: float = 0.0,
        validation_data: Optional[Tuple[np.ndarray, np.ndarray]] = None,
        shuffle: bool = True,
        class_weight: Dict[int, float] = None,
        sample_weight: np.ndarray = None,
        initial_epoch: int = 0,
        steps_per_epoch: int = None,
        validation_steps: int = None,
        validation_freq: int = 1,
        # Note: Max_queue_size, workers and use_multiprocessing not supported as we force input to be numpy array
        pickle_arrays=True,
        num_processes=None,
        context: Callable[[int], EnsembleContextManager] = None,
    ):
        """
        An easy access to keras fit function.
        As the inputs are pickled and distributed onto processes, only numpy array are accepted for the data params
        and no callbacks can be provided.

        If this is too restrictive for your use-case, consider using model.modify to setup your fit process
        and generate the datasets / callbacks right in the map_function.

        :param x: See `tf.keras.Model.fit` documentation.
        :param y: See `tf.keras.Model.fit` documentation.
        :param batch_size: See `tf.keras.Model.fit` documentation.
        :param epochs: See `tf.keras.Model.fit` documentation.
        :param verbose: See `tf.keras.Model.fit` documentation.
        :param callbacks: See `tf.keras.Model.fit` documentation.
        :param validation_split: See `tf.keras.Model.fit` documentation.
        :param validation_data: See `tf.keras.Model.fit` documentation.
        :param shuffle: See `tf.keras.Model.fit` documentation.
        :param class_weight: See `tf.keras.Model.fit` documentation.
        :param sample_weight: See `tf.keras.Model.fit` documentation.
        :param initial_epoch: See `tf.keras.Model.fit` documentation.
        :param steps_per_epoch: See `tf.keras.Model.fit` documentation.
        :param validation_steps: See `tf.keras.Model.fit` documentation.
        :param validation_freq: See `tf.keras.Model.fit` documentation.
        :param pickle_arrays: If true, the arrays are stored to the file system
            and deserialized in every child process to save memory.
        :param num_processes: The number of processes to use.
            Default: The default or value specified when creating the lazy ensemble.
        :param context: A contextmanager which prepares a newly crated process for execution
            (e.g. by configuring the gpus). See class docstring for explanation of default values.
        :return: The fit histories of the atomic models
        """
        if callbacks is not None and callbacks != []:
            raise ValueError(
                "Callbacks are currently not supported in the _lazy_ensemble fit function."
                "Use model.modify to setup your fit process with custom callbacks."
            )
        invalid_input = None
        if not isinstance(x, np.ndarray):
            invalid_input = "x_train"
        elif not isinstance(y, np.ndarray):
            invalid_input = "y_train"
        elif validation_data is not None and (
            validation_data[0] is None or not isinstance(validation_data[0], np.ndarray)
        ):
            invalid_input = "validation_data[0]"
        elif validation_data is not None and (
            validation_data[1] is None or not isinstance(validation_data[1], np.ndarray)
        ):
            invalid_input = "validation_data[1]"
        if invalid_input is not None:
            raise TypeError(
                f"{invalid_input} is not a numpy array. "
                "Lazy Ensembles fit function supports only numpy array, as other types of inputs "
                "may not be efficiently picklable. "
                "Use model.modify to setup your fit process if you want to use other input types"
                "(such as tf.data.Datasets) and generate the dataset in the custom process "
                "to avoid serialization & deserialization."
            )

        x_path = None
        y_path = None
        val_paths = None
        if pickle_arrays:
            # Temporarily save arrays to file system
            x_path = _store_to_temp_file(x)
            y_path = _store_to_temp_file(y)
            if validation_data is not None:
                val_x_path = _store_to_temp_file(validation_data[0])
                val_y_path = _store_to_temp_file(validation_data[1])
                val_paths = (val_x_path, val_y_path)

        inner_fit_function = NumpyFitProcess(
            x=x_path if pickle_arrays else x,
            y=y_path if pickle_arrays else y,
            batch_size=batch_size,
            epochs=epochs,
            verbose=verbose,
            validation_split=validation_split,
            validation_data=val_paths if pickle_arrays else validation_data,
            shuffle=shuffle,
            class_weight=class_weight,
            sample_weight=sample_weight,
            initial_epoch=initial_epoch,
            steps_per_epoch=steps_per_epoch,
            validation_steps=validation_steps,
            validation_freq=validation_freq,
        )
        res = self.modify(
            map_function=inner_fit_function,
            num_processes=num_processes,
            context=context,
        )

        if pickle_arrays:
            # Temporarily save arrays to file system
            os.remove(x_path)
            os.remove(y_path)
            if validation_data is not None:
                os.remove(val_paths[0])
                os.remove(val_paths[1])

        return res

    def predict_quantified(
        self,
        x: np.ndarray,
        quantifier: Union[Quantifier, Iterable[Union[str, Quantifier]]],
        # Other Sequential.predict params (e.g. Callbacks) are not yet supported
        batch_size: int = 32,
        verbose: int = 0,
        steps=None,
        as_confidence: Union[None, bool] = None,
        num_processes=None,
        context=None,
    ):
        """
        Utility function to make quantified predictions on numpy arrays.
        Note: The numpy arrays are replicated on every created process and will thus quickly consume a lot of memory.

        :param x: An (unbatched) numpy array, to be used in tf.keras.Model.predict
        :param quantifier: A single or a collection of (sampling expecting) uwiz.quantifiers
        :param batch_size: The batch size to use in tf.keras.Model.predict
        :param verbose: Not yet supported.
        :param steps: The number of steps to use in tf.keras.Model.predict
        :param as_confidence: If true, uncertainties are multiplied by (-1), if false,
            confidences are multiplied by (-1). Default: No transformations.
        :param num_processes: The number of processes to use. Default:
            The default or value specified when creating the lazy ensemble.
        :param context: A contextmanager which prepares a newly crated process for execution
            (e.g. by configuring the gpus). See class docstring for explanation of default values.
        :return: A tuple (predictions, uncertainties_or_confidences) if a single quantifier was passed as string
            or instance, or a collection of such tuples if the passed quantifiers was an iterable.
        """
        if verbose > 0:
            warnings.warn("Verbosity not yet supported in lazy ensemble models.")
        data_loaded_predictor = DataLoadedPredictor(
            x_test=x, batch_size=batch_size, steps=steps
        )
        return self.quantify_predictions(
            quantifier=quantifier,
            consume_function=data_loaded_predictor,
            as_confidence=as_confidence,
            num_processes=num_processes,
            context=context,
        )

    def quantify_predictions(
        self,
        quantifier: Union[Quantifier, Iterable[Quantifier]],
        consume_function: Callable[[int, tf.keras.Model], Any],
        as_confidence: bool = None,
        num_processes: int = None,
        context: Callable[[int], EnsembleContextManager] = None,
    ):
        """
        A utility function to make predictions on all atomic models and then infer overall predictions and uncertainty
        (or confidence) on those predictions.

        The test data is expected to be loaded directly in the consume function.
        This function, which gets the atomic model id and the atomic model as inputs,
        is expected to return the predictions, i.e., the results of a model.predict(..) call.
        :param quantifier: A single or a collection of (sampling expecting) uwiz.quantifiers
        :param consume_function: A picklable function to make predictions on atomic models, as explained in the description above.
        :param as_confidence: If true, uncertainties are multiplied by (-1), if false, confidences are multiplied by (-1). Default: No transformations.
        :param num_processes: The number of processes to use. Default: The default or value specified when creating the lazy ensemble.
        :param context: A contextmanager which prepares a newly crated process for execution (e.g. by configuring the gpus). See class docstring for explanation of default values.
        :return: A tuple (predictions, uncertainties_or_confidences) if a single quantifier was passed as string or instance, or a collection of such tuples if the passed quantifiers was an iterable.
        """
        all_q, pp_q, sample_q, return_single_tuple = self._quantifiers_as_list(
            quantifier
        )
        self._check_quantifier_heterogenity(
            as_confidence=as_confidence, quantifiers=all_q
        )

        nn_outputs_by_model = self.consume(
            consume_function=consume_function,
            num_processes=num_processes,
            context=context,
        )

        scores = None
        for i, predictions in enumerate(nn_outputs_by_model):
            if scores is None:
                scores_shape = list(predictions.shape)
                scores_shape.insert(1, self.num_models)
                scores = np.empty(scores_shape)
            scores[:, i] = predictions

        results = []
        for q in all_q:
            predictions, superv_scores = q.calculate(scores)
            superv_scores = q.cast_conf_or_unc(
                as_confidence=as_confidence, superv_scores=superv_scores
            )
            results.append((predictions, superv_scores))
        if return_single_tuple:
            return results[0]
        return results

    def _run_in_processes(
        self,
        process_creator,
        inner_function: Callable[[int, tf.keras.Model], Any],
        num_processes: Optional[int],
        context: Optional[Callable[[int], EnsembleContextManager]],
    ):
        num_processes = self._num_processes_or_default(num_processes=num_processes)
        if num_processes > 0:
            if context is None:
                # As default, we just grow the memory dynamically
                context = DynamicGpuGrowthContextManager
            multiprocess_spawn_ctx = multiprocessing.get_context("spawn")
            context.before_start()
            tasks_per_child = context.max_sequential_tasks_per_process()
            pool = multiprocessing.pool.Pool(
                num_processes,
                maxtasksperchild=tasks_per_child,
                context=multiprocess_spawn_ctx,
            )
            partial_process_creator = partial(
                process_creator,
                save_config=self.save_config,
                inner_func=inner_function,
                context=context,
            )
            results = pool.map(partial_process_creator, range(self.num_models))
            pool.close()
            pool.join()
            context.after_end()
            return results
        else:
            if context is None:
                context = NoneContextManager
            res = []
            context.before_start()
            for i in range(self.num_models):
                print(
                    f"Working on model {i} (of {self.num_models}) in the main process"
                )
                model_res = process_creator(
                    model_id=i,
                    save_config=self.save_config,
                    inner_func=inner_function,
                    context=context,
                )
                res.append(model_res)
            context.after_end()
            return res



from typing import List

import numpy as np

import uncertainty_wizard as uwiz

from .quantifier import ConfidenceQuantifier, ProblemType, UncertaintyQuantifier


def _check_inputs_array(inputs, quantifier_name):
    """
    Checks if the input array has the right shape and properties.
    Returns the array as well as a flag indicating whether the input was batched or a single sample
    """
    assert inputs.ndim == 2, (
        "The input to calculate {0} must be two"
        "dimensional (num_inputs x num_classes)".format(quantifier_name)
    )

    # Check that all values are between 0 and 1
    assert np.all(inputs >= 0) and np.all(inputs <= 1), (
        "{0} is built on softmax outputs, but the input array does not represent softmax outputs: "
        "There are entries which are not in the interval [0,1]".format(quantifier_name)
    )

    # Check that all softmax values sum up to one
    assert np.all(
        np.isclose(np.sum(inputs, axis=1), np.ones(shape=inputs.shape[0]), 0.00001)
    ), (
        "{0} is built on softmax outputs, but the input array does not represent softmax outputs: "
        "The per-sample values do not sum up to 1".format(quantifier_name)
    )


class PredictionConfidenceScore(ConfidenceQuantifier):
    """
    The Prediction Confidence Score is a confidence metric in one-shot classification.
    Inputs/activations have to be normalized using the softmax function over all classes.
    The class with the highest activation is chosen as prediction,
    the difference between the two highest activations is used as confidence quantification.
    """

    # docstr-coverage:inherited
    @classmethod
    def takes_samples(cls) -> bool:
        return False

    # docstr-coverage:inherited
    @classmethod
    def problem_type(cls) -> ProblemType:
        return ProblemType.CLASSIFICATION

    # docstr-coverage:inherited
    @classmethod
    def aliases(cls) -> List[str]:
        return ["pcs", "prediction_confidence_score", "PredictionConfidenceScore"]

    # docstr-coverage:inherited
    @classmethod
    def calculate(cls, nn_outputs: np.ndarray):
        _check_inputs_array(nn_outputs, quantifier_name="prediction_confidence_score")

        num_samples = nn_outputs.shape[0]
        calculated_predictions = np.argmax(nn_outputs, axis=1)
        max_values = nn_outputs[np.arange(num_samples), calculated_predictions]
        values_copy = nn_outputs.copy()
        values_copy[np.arange(num_samples), calculated_predictions] = -np.inf
        second_highest_values = np.max(values_copy, axis=1)

        pcs = max_values - second_highest_values
        return calculated_predictions, pcs


class MaxSoftmax(ConfidenceQuantifier):
    """
    The MaxSoftmax is a confidence metric in one-shot classification.
    It is the defaults in most simple use cases and sometimes also referred to
    as 'Vanilla Confidence Metric'.

    Inputs/activations have to be normalized using the softmax function over all classes.
    The class with the highest activation is chosen as prediction,
    the activation of this highest activation is used as confidence quantification.
    """

    # docstr-coverage:inherited
    @classmethod
    def aliases(cls) -> List[str]:
        return ["softmax", "MaxSoftmax", "max_softmax", "sm"]

    # docstr-coverage:inherited
    @classmethod
    def takes_samples(cls) -> bool:
        return False

    # docstr-coverage:inherited
    @classmethod
    def problem_type(cls) -> ProblemType:
        return ProblemType.CLASSIFICATION

    # docstr-coverage:inherited
    @classmethod
    def calculate(cls, nn_outputs: np.ndarray):
        _check_inputs_array(nn_outputs, quantifier_name="softmax")

        num_samples = nn_outputs.shape[0]
        calculated_predictions = np.argmax(nn_outputs, axis=1)
        max_values = nn_outputs[np.arange(num_samples), calculated_predictions]

        return calculated_predictions, max_values


class SoftmaxEntropy(UncertaintyQuantifier):
    """
    The SoftmaxEntropy is a confidence metric in one-shot classification.

    Inputs/activations have to be normalized using the softmax function over all classes.
    The class with the highest activation is chosen as prediction,
    the entropy over all activations is used as uncertainty quantification.
    """

    # docstr-coverage:inherited
    @classmethod
    def aliases(cls) -> List[str]:
        return ["softmax_entropy", "SoftmaxEntropy", "se"]

    # docstr-coverage:inherited
    @classmethod
    def takes_samples(cls) -> bool:
        return False

    # docstr-coverage:inherited
    @classmethod
    def problem_type(cls) -> ProblemType:
        return ProblemType.CLASSIFICATION

    # docstr-coverage:inherited
    @classmethod
    def calculate(cls, nn_outputs: np.ndarray):
        _check_inputs_array(nn_outputs, quantifier_name="softmax-entropy")

        calculated_predictions = np.argmax(nn_outputs, axis=1)
        entropies = uwiz.quantifiers.predictive_entropy.entropy(nn_outputs, axis=1)

        return calculated_predictions, entropies



from .mean_softmax import MeanSoftmax
from .mutual_information import MutualInformation
from .one_shot_classifiers import MaxSoftmax, PredictionConfidenceScore, SoftmaxEntropy
from .predictive_entropy import PredictiveEntropy
from .quantifier import Quantifier
from .regression_quantifiers import StandardDeviation
from .variation_ratio import VariationRatio


class QuantifierRegistry:
    """
    The quantifier registry keeps track of all quantifiers and their string aliases.
    This is primarily used to allow to pass string representations of quantifiers in predict_quantified
    method calls, but may also be used for other purposes where dynamic quantifier selection is desired.
    """

    _registries = dict()

    @classmethod
    def register(cls, quantifier: Quantifier) -> None:
        """
        Use this method to add a new quantifier to the registry.
        :param quantifier: The quantifier instance to be added.
        :return: None
        """
        for alias in quantifier.aliases():
            if alias.lower() in cls._registries:
                raise ValueError(
                    f"A quantifier with alias '{alias}' is already registered."
                )
            cls._registries[alias.lower()] = quantifier

    @classmethod
    def find(cls, alias: str) -> Quantifier:
        """
        Find quantifiers by their id.
        :param alias: A string representation of the quantifier, as defined in the quantifiers aliases method
        :return: A quantifier instance
        """
        record = cls._registries.get(alias.lower())
        if record is None:
            raise ValueError(
                f"No quantifier with alias '{alias}' was found. Check if you made any typos."
                f"If you use the alias of a custom quantifier (i.e., not an uwiz default quantifier),"
                f"make sure to register it through `uwiz.QuantifierRegistry.register(...)`"
            )
        return record


# Register uwiz classification quantifiers
QuantifierRegistry.register(MaxSoftmax())
QuantifierRegistry.register(PredictionConfidenceScore())
QuantifierRegistry.register(SoftmaxEntropy())
QuantifierRegistry.register(VariationRatio())
QuantifierRegistry.register(PredictiveEntropy())
QuantifierRegistry.register(MutualInformation())
QuantifierRegistry.register(MeanSoftmax())

# Register uwiz classification quantifiers
QuantifierRegistry.register(StandardDeviation())



import shutil

import tensorflow as tf

from uncertainty_wizard.models._stochastic._stochastic_mode import StochasticMode

from ._abstract_stochastic import Stochastic
from ._functional_stochastic import StochasticFunctional
from ._sequential_stochastic import StochasticSequential


def stochastic_from_keras(
    model: tf.keras.models.Model,
    input_tensors=None,
    clone_function=None,
    expect_determinism=False,
    temp_weights_path="tmp/weights",
):
    """
    Creates a stochastic instance from a given `tf.keras.models.Sequential` model:
    The new model will have the same structure (layers) and weights as the passed model.

    All stochastic layers (e.g. tf.keras.layers.Dropout) will be used for randomization during randomized predictions.
    If no stochastic layers are present, a ValueError is thrown.
    The raising of the error can be suppressed by setting `expect_determinism` to true.

    :param model: The model to copy. Remains unchanged.
    :param input_tensors: Optional tensors to use as input_tensors for new model. See the corresponding parameter in `tf.keras.models.clone_model` for details.
    :param _clone_function: Optional function to use to clone layers. Will be applied to all layers except input layers and stochastic layers. See the corresponding parameter in `tf.keras.models.clone_model` for more details.
    :param expect_determinism: If True, deterministic models (e.g. models without stochastic layers) are accepted and no ValueError is thrown.
    :param temp_weights_path: The model weights are temporarily saved to the disk at this path. Folder is deleted after successful completion.
    :return: A newly created stochastic model
    """
    # _clone_function is some layer cloning behavior that can be specified by the user
    # If none is specified, we use keras default (see `tf.keras.models.clone_model`)
    if clone_function is None:

        def _clone_function(layer):
            return layer.__class__.from_config(layer.get_config())

    # We wrap the users (or default) clone function in a clone function
    #   that replaces stochastic layers with uncertainty wizard stochastic layers
    stochastic_mode = StochasticMode()
    is_stochastic_layer = []

    def _uncertainty_wizard_aware_clone_function(layer):
        new_layer = Stochastic._replace_layer_if_possible(
            layer, stochastic_mode=stochastic_mode
        )
        if new_layer == layer:
            # Layer was not mapped to an uncertainty wizard layer, thus the default clone function is applied
            new_layer = _clone_function(layer)
            is_stochastic_layer.append(False)
        else:
            is_stochastic_layer.append(True)
        return new_layer

    # Clone the keras model to become the new inner model
    new_inner = tf.keras.models.clone_model(
        model=model,
        input_tensors=input_tensors,
        clone_function=_uncertainty_wizard_aware_clone_function,
    )
    new_inner.stochastic_mode_tensor = stochastic_mode.as_tensor()

    if not expect_determinism and not any(is_stochastic_layer):
        raise ValueError(
            "The passed model had no stochastic layers."
            "If that is intended (and you do not plan to use any sampling based quantifiers)"
            "you can set the flag `expect_determinism = True`, i.e., "
            "calling `SequentialStochastic.clone_from_keras(keras_model,expect_determinism = True)`"
        )

    # Restore the Weights
    model.save_weights(temp_weights_path)
    new_inner.load_weights(temp_weights_path)
    # Remove temporarily stored weights
    shutil.rmtree(temp_weights_path, ignore_errors=True)

    # Put the wrapper around the new model
    if isinstance(model, tf.keras.models.Sequential):
        target_class = StochasticSequential
    else:
        target_class = StochasticFunctional

    # Consenting Adults: The _wrap method is intended to be used here
    #   but declared private as it is not intended to be used by the uwiz user
    return target_class._wrap(
        inner=new_inner, stochastic_mode_tensor=stochastic_mode.as_tensor()
    )



"""This module contains all quantifiers used to infer prediction and confidence (or uncertainty)
from neural network outputs. It also contains the QuantifierRegistry which allows to refer to
quantifiers by alias."""

__all__ = [
    "Quantifier",
    "MutualInformation",
    "PredictiveEntropy",
    "VariationRatio",
    "MeanSoftmax",
    "PredictionConfidenceScore",
    "MaxSoftmax",
    "SoftmaxEntropy",
    "StandardDeviation",
    "QuantifierRegistry",
]


# Base class
from .mean_softmax import MeanSoftmax

# Sampling Based Classification Quantifiers
from .mutual_information import MutualInformation

# Point Predictor Classification Quantifiers
from .one_shot_classifiers import MaxSoftmax, PredictionConfidenceScore, SoftmaxEntropy
from .predictive_entropy import PredictiveEntropy
from .quantifier import Quantifier

# Registry
from .quantifier_registry import QuantifierRegistry

# Regression
from .regression_quantifiers import StandardDeviation
from .variation_ratio import VariationRatio



"""
The layers in this file are extensions of the randomized keras layers,
which are modified in a way to take the stochastic mode into account.
"""
import warnings

import tensorflow as tf
from tensorflow.python.keras.layers import GaussianNoise

from uncertainty_wizard.internal_utils import UncertaintyWizardWarning
from uncertainty_wizard.models._stochastic._stochastic_mode import StochasticMode

_MISSING_STOCHASTIC_MODE_ERROR = (
    "No stochastic mode instance was provided when creating the randomized layer. "
    "A stochastic mode is required to use the randomization for predictions"
)


def _has_casting_preventing_subtype(
    instance, expected_type, corresponding_uw_type
) -> bool:
    i_type = type(instance)
    if issubclass(i_type, expected_type) and not issubclass(expected_type, i_type):
        # The instance is from a (strict) subclass of the expected type
        if isinstance(instance, corresponding_uw_type):
            warnings.warn(
                f"Looks like you are passing an {corresponding_uw_type} layer."
                f"For SequentialStochastic layers, it is sufficient to pass a layer of"
                f"the corresponding keras layer {expected_type}."
                f"We trust you that you know what you did and set up the stochastic mode correctly."
                f"Your layer will thus not be replaced, but added to the model as you provided it.",
                UncertaintyWizardWarning,
            )
        else:
            warnings.warn(
                f"Looks like you are passing an instance of a custom subtype of {expected_type}."
                f"We typically replace {expected_type} instances with our own custom subtype."
                f"We will not do this with your custom subtype instance. "
                f"If you want to use it for randomness during inference, "
                f"make sure the models stochastic mode tensor is respected in your layer.",
                UncertaintyWizardWarning,
            )
        return True
    return False


class UwizBernoulliDropout(tf.keras.layers.Dropout):
    """
    The extension of tf.keras.layers.Dropout to be used in uncertainty wizard stochastic models
    """

    def __init__(
        self,
        rate,
        stochastic_mode: StochasticMode,
        noise_shape=None,
        seed=None,
        **kwargs,
    ):
        """
        Create a new layer instance. This is essentially the same as creating a tf.keras.layers.Dropout layer,
        but in addition, a stochastic mode is expected, which will allow to dynamically toggle randomness at runtime.
        :param rate: see the corresponding keras docs.
        :param stochastic_mode: A stochastic mode instance. Must be the same thats used in the functional model this layer will be added to.
        :param noise_shape: see the corresponding keras docs.
        :param seed: see the corresponding keras docs.
        :param kwargs: see the corresponding keras docs.
        """
        super().__init__(rate, noise_shape, seed, **kwargs)
        assert stochastic_mode is not None, _MISSING_STOCHASTIC_MODE_ERROR
        self.stochastic_mode = stochastic_mode
        # We must not call super from within tf.cond
        self._super = super()

    # docstr-coverage:inherited
    def call(self, inputs, training=None):
        return tf.cond(
            self.stochastic_mode.as_tensor(),
            lambda: self._super.call(inputs=inputs, training=True),
            lambda: self._super.call(inputs=inputs, training=training),
        )

    @classmethod
    def from_keras_layer(
        cls, layer: tf.keras.layers.Dropout, stochastic_mode: StochasticMode
    ):
        """
        Attempts to create a new UwizBernoulliDropout instance based on the configuration (i.e. dropout rate)
        of a passed Dropout layer
        :param layer: The layer from which to read the dropout layer
        :param stochastic_mode: The stochastic mode which allows to toggle randomness.
        :return: A UwizBernoulliDropout, if casting was successful. Otherwise (i.e., if the passed layer was a casting preventing subtype of Dropout), the passed layer is returned and a warning is printed to the console.
        """
        if _has_casting_preventing_subtype(
            layer, tf.keras.layers.Dropout, UwizBernoulliDropout
        ):
            return layer
        else:
            rate = layer.rate
            noise_shape = layer.noise_shape
            seed = layer.seed
            return UwizBernoulliDropout(
                rate=rate,
                noise_shape=noise_shape,
                seed=seed,
                stochastic_mode=stochastic_mode,
            )

    # docstr-coverage:inherited
    def get_config(self):
        config = super(UwizBernoulliDropout, self).get_config()
        config["name"] = "UwBernoulliDropout"
        return config


class UwizGaussianDropout(tf.keras.layers.GaussianDropout):
    """
    The extension of tf.keras.layers.GaussianDropout to be used in uncertainty wizard stochastic models
    """

    def __init__(self, rate, stochastic_mode: StochasticMode, **kwargs):
        """
        Create a new layer instance. This is essentially the same as creating a tf.keras.layers.Dropout layer,
        but in addition, a stochastic mode is expected, which will allow to dynamically toggle randomness at runtime.
        :param rate: see the corresponding keras docs.
        :param stochastic_mode: A stochastic mode instance. Must be the same which is going to be used in the functional model this layer will be added to.
        :param kwargs: see the corresponding keras docs.
        """
        super().__init__(rate, **kwargs)
        assert stochastic_mode is not None, _MISSING_STOCHASTIC_MODE_ERROR
        self.stochastic_mode = stochastic_mode
        # We must not call super from within tf.cond
        self._super = super()

    # docstr-coverage:inherited
    def call(self, inputs, training=None):
        return tf.cond(
            self.stochastic_mode.as_tensor(),
            lambda: self._super.call(inputs=inputs, training=True),
            lambda: self._super.call(inputs=inputs, training=training),
        )

    @classmethod
    def from_keras_layer(
        cls, layer: tf.keras.layers.GaussianDropout, stochastic_mode: StochasticMode
    ):
        """
        Attempts to create a new UwizGaussianDropout instance based on the configuration (i.e. dropout rate)
        of a passed GaussianDropout layer
        :param layer: The layer from which to read the dropout layer
        :param stochastic_mode: The stochastic mode which allows to toggle randomness.
        :return: A UwizGaussianDropout, if casting was successful. Otherwise (i.e., if the passed layer was a casting preventing subtype of GaussianDropout), the passed layer is returned and a warning is printed to the console.
        """
        if _has_casting_preventing_subtype(
            layer, tf.keras.layers.GaussianDropout, UwizGaussianDropout
        ):
            return layer
        else:
            rate = layer.rate
            return UwizGaussianDropout(rate=rate, stochastic_mode=stochastic_mode)

    # docstr-coverage:inherited
    def get_config(self):
        config = super(UwizGaussianDropout, self).get_config()
        config["name"] = "UwGaussianDropout"
        # No custom config yet.
        return config


class UwizGaussianNoise(tf.keras.layers.GaussianNoise):
    """
    The extension of tf.keras.layers.GaussianNoise to be used in uncertainty wizard stochastic models
    """

    def __init__(self, stddev, stochastic_mode: StochasticMode, **kwargs):
        """
        Create a new layer instance. This is essentially the same as creating a tf.keras.layers.Dropout layer,
        but in addition, a stochastic mode is expected, which will allow to dynamically toggle randomness at runtime.
        :param stddev: see the corresponding keras docs.
        :param stochastic_mode: A stochastic mode instance. Must be the same thats used in the functional model this layer will be added to.
        :param kwargs: see the corresponding keras docs.
        """
        super().__init__(stddev, **kwargs)
        assert stochastic_mode is not None, _MISSING_STOCHASTIC_MODE_ERROR
        self.stochastic_mode = stochastic_mode
        # We must not call super from within tf.cond
        self._super = super()

    # docstr-coverage:inherited
    def call(self, inputs, training=None):
        return tf.cond(
            self.stochastic_mode.as_tensor(),
            lambda: self._super.call(inputs=inputs, training=True),
            lambda: self._super.call(inputs=inputs, training=training),
        )

    @classmethod
    def from_keras_layer(
        cls, layer: tf.keras.layers.GaussianNoise, stochastic_mode: StochasticMode
    ):
        """
        Attempts to create a new UwizGaussianNoise instance based on the configuration (i.e. the standard deviation)
        of a passed GaussianNoise layer
        :param layer: The layer from which to read the dropout layer
        :param stochastic_mode: The stochastic mode which allows to toggle randomness.
        :return: A UwizGaussianNoise, if casting was successful. Otherwise (i.e., if the passed layer was a casting preventing subtype of GaussianNoise), the passed layer is returned and a warning is printed to the console.
        """
        if _has_casting_preventing_subtype(
            layer, tf.keras.layers.GaussianNoise, UwizGaussianNoise
        ):
            return layer
        else:
            stddev = layer.stddev
            return UwizGaussianNoise(stddev=stddev, stochastic_mode=stochastic_mode)

    # docstr-coverage:inherited
    def get_config(self):
        config = super(GaussianNoise, self).get_config()
        config["name"] = "UwGaussianNoise"
        # No custom config yet.
        return config



import tensorflow as tf

from uncertainty_wizard.models._stochastic._abstract_stochastic import Stochastic
from uncertainty_wizard.models._stochastic._stochastic_mode import StochasticMode


class StochasticFunctional(Stochastic):
    """
    A stochastic wrapper of a `tf.keras.Model`, allowing to build models using the functional interface.
    Note that when using the functional interface, you need to use `uwiz.models.stochastic.layers`
    or build your own Stochastic-Mode dependent stochastic layers. See the online user guide for more info.

    """

    # Include the superclass documentation
    __doc__ += Stochastic.__doc__

    def __init__(
        self, inputs, outputs, stochastic_mode: StochasticMode, name: str = None
    ):
        """
        Create a new functional model, equivalent to calling tf.keras.Model(...).

        In addition, a stochastic mode instance has to be passed.
        The same instance also has to be passed to any randomized uwiz.layers instances
        which are part of this model.
        This allows to dynamically enable and disable randomness in the predictions.
        :param inputs: See the corresponding tf.keras.Model(...) docs
        :param outputs: See the corresponding tf.keras.Model(...) docs
        :param stochastic_mode: A stochastic mode instance
        :param name: See the corresponding tf.keras.Model(...) docs
        """
        super().__init__()
        self._inner_model = tf.keras.Model(inputs=inputs, outputs=outputs, name=name)
        self._inner_model._stochastic_mode_tensor = stochastic_mode.as_tensor()

    # docstr-coverage:inherited
    @property
    def inner(self):
        return self._inner_model

    # docstr-coverage:inherited
    @property
    def stochastic_mode_tensor(self):
        return self._inner_model._stochastic_mode_tensor

    @classmethod
    def _wrap(cls, inner: tf.keras.Model, stochastic_mode_tensor=None):
        if stochastic_mode_tensor is None:
            assert inner._stochastic_mode_tensor is not None, (
                "Uncertainty Wizard internal error. "
                "Trying to wrap a model that has no stochastic_mode_tensor, "
                "and no external stochastic_mode_tensor is passed to attach"
            )
            stochastic_mode_tensor = inner._stochastic_mode_tensor
        stochastic_mode = StochasticMode(stochastic_mode_tensor)
        return StochasticFunctional(
            inner.inputs, inner.outputs, stochastic_mode=stochastic_mode
        )



import json
import logging
import os
import warnings

import tensorflow as tf

import uncertainty_wizard as uwiz
from uncertainty_wizard.internal_utils import (
    UncertaintyWizardWarning,
    tf_version_resolver,
)

from ._stochastic._functional_stochastic import StochasticFunctional
from ._stochastic._sequential_stochastic import StochasticSequential


def load_model(path, custom_objects: dict = None, compile=None, options=None):
    """
    Loads an uncertainty wizard model that was saved using model.save(...).
    See the documentation of `tf.keras.models.load_model` for further information about the method params.

    For lazy ensembles: As they are lazy, only the folder path and the number of models are interpreted
    by this model loading - no keras models are actually loaded yet.
    Thus, custom_objects, compile and options must not be specified.

    :param path: The path of the folder where the ensemble was saved.
    :param custom_objects: Dict containing methods for custom deserialization of objects.
    :param compile: Whether to compile the models.
    :param options: Load options, check tf.keras documentation for precise information.

    :return: An uwiz model.
    """
    # Note: a path without file extension is not necessarily an ensemble in a folder
    # It could be a user who stored a stochastic model with default (non specified) file ending
    # Thus, we have to check if the folder exists and contains an ensemble config file
    ensemble_config_path = uwiz.models.ensemble_utils._lazy_ensemble.config_file_path(
        path
    )
    if (
        os.path.isdir(path)
        and os.path.exists(path)
        and os.path.exists(ensemble_config_path)
    ):
        return _load_ensemble(
            path=path, custom_objects=custom_objects, compile=compile, options=options
        )
    else:
        return _load_stochastic(
            path=path, custom_objects=custom_objects, compile=compile, options=options
        )


def _load_stochastic(path, custom_objects: dict = None, compile=None, options=None):
    """Attempts to load the model at the provided path as a stochastic model"""

    # Note: We currently intentionally don't define stochastic layers as custom_objects
    # as they have no methods other than call that we rely on, and thus the (robust and easy to maintain)
    # tf deserialization is sufficient
    if tf_version_resolver.current_tf_version_is_older_than("2.3.0", fallback=True):
        if options is not None:
            raise ValueError(
                "Load-Options are not supported by tensorflow<2.3.0."
                "Please do not specify any options when you call 'uwiz.models.load_model'"
                "or upgrade to a tensorflow version >= 2.3.0"
            )
        inner = tf.keras.models.load_model(
            path, custom_objects=custom_objects, compile=compile
        )
    else:
        inner = tf.keras.models.load_model(
            path, custom_objects=custom_objects, compile=compile, options=options
        )
    assert hasattr(
        inner, "_stochastic_mode_tensor"
    ), "Looks like the model which is being deserialized is not an uwiz stochastic model"

    if isinstance(inner, tf.keras.models.Sequential):
        return StochasticSequential._wrap(inner)
    else:
        return StochasticFunctional._wrap(inner)


def _load_ensemble(path, custom_objects: dict = None, compile=None, options=None):
    """Creates a lazy ensemble with the provided path as root dir. No models are acutally loaded yet (as in 'lazy')."""

    if compile is not None or options is not None or custom_objects is not None:
        warnings.warn(
            "Parameters compile, custom_objects and options are still ignored in lazy ensembles."
            "Support may be added in the future.",
            UncertaintyWizardWarning,
        )

    with open(
        uwiz.models.ensemble_utils._lazy_ensemble.config_file_path(path=path), "r"
    ) as f:
        config = json.load(f)

    num_models = config["num_models"]
    ensemble = uwiz.models.LazyEnsemble(
        num_models=num_models,
        model_save_path=path,
        expect_model=True,
        delete_existing=False,
    )
    logging.info(
        "Loaded ensemble. You may want to override the default_num_processes 'model.default_num_processes'"
    )
    return ensemble



__all__ = [
    "EnsembleContextManager",
    "DynamicGpuGrowthContextManager",
    "NoneContextManager",
    "DeviceAllocatorContextManager",
    "SaveConfig",
]

from ._lazy_contexts import (
    DeviceAllocatorContextManager,
    DynamicGpuGrowthContextManager,
    EnsembleContextManager,
    NoneContextManager,
)
from ._save_config import SaveConfig



from typing import List, Tuple

import numpy as np

from .quantifier import ProblemType, UncertaintyQuantifier
from .variation_ratio import VariationRatio


def entropy(data: np.ndarray, axis: int) -> np.ndarray:
    """
    A utility method to compute the entropy. May also be used by other quantifiers which internally rely on entropy.

    Following standard convention, the logarithm used in the entropy calculation is on base 2.

    :param data: The values on which the entropy should be computed.
    :param axis: Entropy will be taken along this axis.
    :return: An array containing the entropies. It is one dimension smaller than the passed data (the specified axis was removed).
    """
    # Remove zeros from nn_outputs (to allow to take logs)
    # Note that the actual increment (1e-20) does not matter, as it is multiplied by 0 below
    increments = np.zeros(shape=data.shape, dtype=np.float32)
    indexes_of_zeros = data == 0
    increments[indexes_of_zeros] = 1e-20
    nonzero_data = data + increments

    # These arrays can be quite large and are not used anymore - we free the space for the operations below
    del increments, indexes_of_zeros

    # Calculate and return the entropy
    return -np.sum(data * np.log2(nonzero_data), axis=axis)


class PredictiveEntropy(UncertaintyQuantifier):
    """
    A predictor & uncertainty quantifier, based on multiple samples (e.g. nn outputs) in a classification problem

    The prediction is made using a plurality vote, i.e., the class with the highest value in most samples is selected.
    In the case of a tie, the class with the lowest index is selected.

    The uncertainty is quantified using the predictive entropy;
    the entropy (base 2) of the per-class means of the sampled predictions.
    """

    # docstr-coverage:inherited
    @classmethod
    def aliases(cls) -> List[str]:
        return ["predictive_entropy", "pred_entropy"]

    # docstr-coverage:inherited
    @classmethod
    def takes_samples(cls) -> bool:
        return True

    # docstr-coverage:inherited
    @classmethod
    def problem_type(cls) -> ProblemType:
        return ProblemType.CLASSIFICATION

    # docstr-coverage:inherited
    @classmethod
    def calculate(cls, nn_outputs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        # For simplicity, we let the predictions be calculated by the Variation Ratio Code
        #    accepting a slight overhead from also calculating the actual variation ratio
        predictions, _ = VariationRatio.calculate(nn_outputs)

        # Take means over the samples
        means = np.mean(nn_outputs, axis=1)

        # The predictive entropy is the entropy of the means
        predictive_entropy = entropy(means, axis=1)

        return predictions, predictive_entropy



import setuptools

import version

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="uncertainty-wizard",
    version=version.RELEASE,
    author="Michael Weiss",
    author_email="code@mweiss.ch",
    description="Quick access to uncertainty and confidence of Keras networks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/testingautomated-usi/uncertainty_wizard",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import warnings
from typing import Union

import tensorflow as tf


def _compare_expected_to_current_tf_version(expected_version) -> Union[None, int]:
    """
    Compares the 'x.y.z' version parts of a passed expected version and the actual tensorflow version.
    The result is negative if the expected version is newer, positive if the expected version is older
    and 0 if they are the same.
    If one of the versions cannot be parsed, a warning is triggered and 'None' is returned.
    :param expected_version:
    :return: an int if a comparison was made and None if parsing was impossible
    """
    actual_version = tf.version.VERSION

    # Note: We're currently only considering versions in the format 'x.y.z'
    # (i.e., ignore RCs and multi digit versions - which is probably fine given tfs very fast major release cycles)

    # Inspection disabling reason: We really want to catch all exceptions.
    # noinspection PyBroadException
    try:
        expected_v_splits = [int(v) for v in expected_version[:5].split(".")]
        actual_v_splits = [int(v) for v in actual_version[:5].split(".")]
    except Exception:
        warnings.warn(
            f"One of the version strings '{expected_version}' (requested) "
            f"or '{actual_version}' was not parsed: "
            f"We are trying to use a suitable guess about your tf compatibility and thus,"
            f"you may not actually note any problems."
            f"However, to be safe, please report this issue (with this warning) "
            f"to the uncertainty wizard maintainers.",
            RuntimeWarning,
        )
        return None

    for act, expected in zip(actual_v_splits, expected_v_splits):
        if expected > act:
            return 1
        elif expected < act:
            return -1
    # Version equality
    return 0


def current_tf_version_is_older_than(version: str, fallback: Union[bool, None] = True):
    """
    A method to check whether the loaded tensorflow version is older than a passed version.
    :param version: A tensorflow version string, e.g. '2.3.0'
    :param fallback: If a problem occurs during parsing, the value of fallback will be returned
    :return: True if the used tensorflow version is older than the version specified in the passed string
    """
    comp = _compare_expected_to_current_tf_version(version)
    if comp is None:
        return fallback
    elif comp > 0:
        return True
    else:
        return False


def current_tf_version_is_newer_than(version: str, fallback: Union[bool, None] = False):
    """
    A method to check whether the loaded tensorflow version is younger than a passed version.
    :param version: A tensorflow version string, e.g. '2.3.0'
    :param fallback: If a problem occurs during parsing, the value of fallback will be returned
    :return: True if the used tensorflow version is newer than the version specified in the passed string
    """
    comp = _compare_expected_to_current_tf_version(version)
    if comp is None:
        return fallback
    elif comp >= 0:
        return False
    else:
        return True



import tensorflow as tf

batch_size = 128
num_classes = 10


class StochasticMode:
    """
    Stochastic mode is a wrapper for a bool tensor which serves as flag during inference in an uwiz stochastic model:
    If the flag is True, the inference in randomized. Otherwise, randomization is disabled.

    When creating a StochasticFunctional model, you need to create a new StochasticMode(),
    use it for any of your (custom?) layers that should have a behavior in a stochastic environment
    than in a detererministic environment (for example your own randomization layer).
    """

    def __init__(self, tensor=None):
        """
        Create a new stochastic mode. If not provided, a new flag tensor will be created.
        :param tensor: Pass your own boolean tensorflow Variable. Use is not recommended.
        """
        if tensor is not None:
            self._enabled = tensor
        else:
            self._enabled = tf.Variable(
                initial_value=False, trainable=False, dtype=bool
            )

    def as_tensor(self):
        """
        Get the tensor wrapped by this stochastic mode
        :return: A boolean tensor
        """
        return self._enabled



from typing import List, Tuple

import numpy as np

from uncertainty_wizard.quantifiers.predictive_entropy import PredictiveEntropy, entropy
from uncertainty_wizard.quantifiers.quantifier import ProblemType, UncertaintyQuantifier


class MutualInformation(UncertaintyQuantifier):
    """
    A predictor & uncertainty quantifier, based on multiple samples (e.g. nn outputs) in a classification problem

    The prediction is made using a plurality vote, i.e., the class with the highest value in most samples is selected.
    In the case of a tie, the class with the lowest index is selected.

    The uncertainty is quantified using the mutual information.
    See the docs for a precise explanation of mutual information.
    """

    # docstr-coverage:inherited
    @classmethod
    def takes_samples(cls) -> bool:
        return True

    # docstr-coverage:inherited
    @classmethod
    def problem_type(cls) -> ProblemType:
        return ProblemType.CLASSIFICATION

    # docstr-coverage:inherited
    @classmethod
    def aliases(cls) -> List[str]:
        return ["mutu_info", "mutual_information", "mi"]

    # docstr-coverage:inherited
    @classmethod
    def calculate(cls, nn_outputs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        predictions, pred_entropy = PredictiveEntropy.calculate(nn_outputs)

        entropies = entropy(nn_outputs, axis=2)
        entropy_means = np.mean(entropies, axis=1)

        # The mutual information is the predictive entropy minus the mean of the entropies
        mutual_information = pred_entropy - entropy_means

        return predictions, mutual_information



import abc
import errno
import os
import pickle
import time
from typing import Dict

import tensorflow as tf

from uncertainty_wizard.models.ensemble_utils._save_config import SaveConfig

global number_of_tasks_in_this_process


def _init_num_tasks():
    global number_of_tasks_in_this_process
    number_of_tasks_in_this_process = 0


_init_num_tasks()


class EnsembleContextManager(abc.ABC):
    """
    An abstract superclass of context managers which can be used to instantiate a context
    on a newly created process.

    Note that subclasses may override the constuctor, but they must not add or remove any arguments from it.

    """

    def __init__(self, model_id: int, varargs: dict = None):
        """
        The constructor currently only receives the id of the atomic model for which
        it will have to generate a context.
        Later, to make it easier for custom child classes of EnsembleContextManager,
        a (now still empty) varargs is also passed which may be populated with more information
        in future versions of uncertainty_wizard.
        """
        self.ensemble_id = (model_id,)
        self.varargs = varargs

    def __enter__(self) -> "EnsembleContextManager":
        """
        Will be executed before session the model is executed.
        Must return 'self'
        :return: None
        """
        return self

    def __exit__(self, type, value, traceback) -> None:
        """
        Will be executed before session the model was executed. You can use this for clean up tasks.
        :return: None
        """
        global number_of_tasks_in_this_process
        number_of_tasks_in_this_process += 1

    @classmethod
    def max_sequential_tasks_per_process(cls) -> int:
        """
        This method is used to specify how often a process (i.e., for how many atomic models)
        every single process can be used before being replaced with a new spawn process.
        Extend this method and set the value to something low to prevent process pollution
        (e.g. if there's a memory leak in your process),
        but beware that process creation (which implies tf initialization) is a costly task.

        The method must be pure; every time it is called it should return the same value.

        The method is ignored if using `num_processes = 0`, i.e., when executing in the main process.

        *Default* If not overridden, this method returns `1000` which means that processes is
        infinitely reused in any reasonable ensemble.

        :return: A positive integer, specifying how often a process can be used before being replaced.
        """
        return 1000

    @classmethod
    def before_start(cls) -> None:
        """
        This method is called once whenever a context-aware call is made on an ensemble
        (e.g., create, modify, predict_quantified, ...). Hence, it is *not* called for every atomic model
        individually.

        Typically, this method would be used to setup requirements that the later created EnsembleContextManager
        instances are relying on.

        Default behavior if method not overridden: Nothing is done.
        """

    @classmethod
    def after_end(cls) -> None:
        """
        This method is called at the end of every context-aware call
        (e.g., create, modify, predict_quantified, ...). Hence, it is *not* called for every atomic model
        individually.

        Typically, this method would be used to setup requirements that the later created EnsembleContextManager
        instances are relying on.

        Default behavior if method not overridden: Nothing is done.
        """

    # Inspection disabled as overriding child classes may want to use 'self'
    # noinspection PyMethodMayBeStatic
    def save_single_model(
        self, model_id: int, model: tf.keras.Model, save_config: SaveConfig
    ) -> None:
        """
        This method will be called to store a single atomic model in the ensemble.
        :param model_id: The id of the atomic model.
        Is between 0 and the number of atomic model in the ensemble.
        :param model: The keras model to be saved.
        :param save_config: A save_config instance, providing information about the base path
        of the ensemble
        """
        path = os.path.abspath(save_config.filepath(model_id=model_id))
        model.save(filepath=path)

    # Inspection disabled as overriding child classes may want to use 'self'
    # noinspection PyMethodMayBeStatic
    def load_single_model(
        self, model_id: int, save_config: SaveConfig
    ) -> tf.keras.Model:
        """
        This method will be called to load a single atomic model in the ensemble.
        :param model_id: The id of the atomic model.
        Is between 0 and the number of atomic model in the ensemble.
        :param save_config: A save_config instance, providing information about the base path
        of the ensemble
        :return The loaded keras model.
        """
        tf.keras.backend.clear_session()
        return tf.keras.models.load_model(
            filepath=save_config.filepath(model_id=model_id)
        )


class NoneContextManager(EnsembleContextManager):
    """
    This context manager makes nothing at all,
    i.e., the model will be executed in exactly the state the process was created.

    This for example implies that the tensorflow default GPU configuration will be used.

    It is save to use this ContextManager on an existing processes where there is already a tf.session
    available.
    """


class DynamicGpuGrowthContextManager(EnsembleContextManager):
    """
    This context manager configures tensorflow such that multiple processes can use the main GPU at the same time.
    It is the default in a lazy ensemble multiprocessing environment
    """

    # docstr-coverage:inherited
    def __enter__(self) -> "DynamicGpuGrowthContextManager":
        super().__enter__()
        global number_of_tasks_in_this_process
        if number_of_tasks_in_this_process == 0:
            self.enable_dynamic_gpu_growth()
        return self

    @classmethod
    def enable_dynamic_gpu_growth(cls):
        """
        Configures tensorflow to set memory growth to ture on all GPUs
        :return: None
        """
        if "CUDA_VISIBLE_DEVICES" in os.environ:
            del os.environ["CUDA_VISIBLE_DEVICES"]
        gpus = tf.config.experimental.list_physical_devices("GPU")
        if gpus:
            try:
                # Currently, memory growth needs to be the same across GPUs
                for gpu in gpus:
                    tf.config.experimental.set_memory_growth(gpu, True)
                logical_gpus = tf.config.experimental.list_logical_devices("GPU")
                print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
            except RuntimeError as e:
                # Memory growth must be set before GPUs have been initialized
                print(e)


global device_id


class DeviceAllocatorContextManager(EnsembleContextManager, abc.ABC):
    """
    This context manager configures tensorflow such a user-defined amount of processes for every available gpu
    are started. In addition, running a process on the CPU can be enabled.

    This is an abstract context manager. To use it, one has to subclass it and override (at least)
    the abstract methods.
    """

    # docstr-coverage: inherited
    def __enter__(self) -> "DeviceAllocatorContextManager":
        super().__enter__()
        global number_of_tasks_in_this_process
        global device_id
        if number_of_tasks_in_this_process == 0:
            device_id = self._get_availabilities_and_choose_device()
            if device_id == -1:
                self._use_cpu()
            else:
                self._use_gpu(index=device_id)
        else:
            print(f"Process on device {device_id} is reused. Not initializing devices.")
        return self

    # docstr-coverage: inherited
    def __exit__(self, type, value, traceback) -> None:
        super().__exit__(type, value, traceback)

        global number_of_tasks_in_this_process
        global device_id
        if number_of_tasks_in_this_process == self.max_sequential_tasks_per_process():
            self._release_device(device_id)
        else:
            print(
                f"Not releasing assignment of device {device_id} as process might be reused."
            )

    # docstr-coverage: inherited
    @classmethod
    def before_start(cls):
        path = cls.file_path()
        if os.path.exists(path):
            os.remove(path)
        lock_path = cls._lock_file_path()
        if os.path.exists(lock_path):
            os.remove(lock_path)

    @classmethod
    @abc.abstractmethod
    def file_path(cls) -> str:
        """
        DeviceAllocatorContextManager uses a temporary file to corrdinate device allocation.
        This method should provide the path of that file.

        The folder where the file should be placed must exist.
        The file itself should not exist (if it does, it will be deleted).

        *Attention:* This function must be pure: Repeated calls should always return the same value.

        :return:
        """

    @classmethod
    def _lock_file_path(cls) -> str:
        """
        DeviceAllocatorContextManager uses a lockfile when coordinating device allocation.
        This method can be used to specify the path of the lock file.

        The folder where the lockfile should be placed must exist.
        The lockfile should not exist (if it does, it will be deleted).

        *Attention:* This function must be pure: Repeated calls should always return the same value.

        :return: The path of the lockfile
        """
        return cls.file_path() + ".lock"

    @classmethod
    @abc.abstractmethod
    def run_on_cpu(cls) -> bool:
        """
        Specify whether one process (in addition to the ones specified in 'virtual_devices_per_gpu'
        should run on the CPU.

        Note that even for the processes running on the GPU, some part of the workload (e.g. preprocessing)
        is typically performed on the CPU and may already consume a lot of resources.
        Do not set this to true if you have a weak CPU - your PC may freeze.

        *Attention:* This function must be pure: Repeated calls should always return the same value.

        :return: A flag indicating if the CPU should also dispatch a job to the cpu.
        """

    @classmethod
    @abc.abstractmethod
    def virtual_devices_per_gpu(cls) -> Dict[int, int]:
        """
        A dict mapping every GPU index to the number of processes which should be started concurrently
        on that GPU.

        *Attention:* This function must be pure: Repeated calls should always return the same value.

        :return: A mapping specifying how many processes of this ensemble should run concurrently per gpu.
        """

    @classmethod
    @abc.abstractmethod
    def gpu_memory_limit(cls) -> int:
        """
        Override this method to specify the amount of MB which should be used
        when creating the virtual device on the GPU. Ignored for CPUs.

        *Attention:* This function must be pure: Repeated calls should always return the same value.

        :return: The amount of MB which will be reserved on the selected gpu in the created context.
        """

    @classmethod
    def acquire_lock_timeout(cls) -> int:
        """
        DeviceAllocatorContextManager uses a lockfile when coordinating device allocation.

        This method can be used to specify a timeout. If the lock is not acquired within this time,
        the process fails.

        :return: The number of seconds to wait until timing out.
        """
        return 5

    @classmethod
    def delay(cls) -> float:
        """
        DeviceAllocatorContextManager uses a lockfile when coordinating device allocation.
        If a file is locked, DeviceAllocatorContextManager tries periodically in the
        interval specified in this method to gain the lock.

        :return: The number of seconds to wait before a retry if the lock cannot be acquired.
        """
        return 0.1

    @classmethod
    def _get_availabilities_and_choose_device(cls) -> int:
        lockfile = cls._acquire_lock()
        try:
            with open(cls.file_path(), "rb") as file:
                availabilities = pickle.load(file)
                device = cls._pick_device(availabilities)

        except OSError as e:
            if not isinstance(e, FileNotFoundError):
                # Some other error than 'file exists' occurred
                raise RuntimeError(
                    (
                        "An error occurred when trying read current allocation file "
                        "for the Uncertainty Wizard DeviceAllocatorContextManager"
                    )
                ) from e

            # All good, this process is apparently the first one and has to create the file.
            availabilities = cls.virtual_devices_per_gpu()
            if cls.run_on_cpu():
                availabilities[-1] = 1
            else:
                availabilities[-1] = 0
            device = cls._pick_device(availabilities)

        availabilities[device] = availabilities[device] - 1
        print(
            f"\nupdating availabilities file: {availabilities} (decreased availability of device {device})"
        )
        with open(cls.file_path(), "wb") as file:
            pickle.dump(availabilities, file)

        cls._release_lock(lockfile=lockfile)
        return device

    @classmethod
    def _release_device(cls, device: int) -> None:
        with open(cls.file_path(), "rb") as file:
            availabilities = pickle.load(file)
        availabilities[device] = availabilities[device] + 1
        print(
            f"\nupdating availabilities file: {availabilities} (increased availability of device {device})"
        )
        with open(cls.file_path(), "wb") as file:
            pickle.dump(availabilities, file)

    @classmethod
    def _pick_device(cls, availablilities) -> int:
        picked_device = None
        picked_device_availability = 0
        for device, availability in availablilities.items():
            if availability > picked_device_availability:
                picked_device = device
                picked_device_availability = availability
        if picked_device is None:
            raise ValueError(
                f"No available devices. Please make sure that the number processes configured "
                f"in your call to the ensemble does not exceed the number of deviced specified "
                f"in your DeviceAllocatorContextManager.virtual_devices_per_gpu extension "
                f"(+1 if run_on_cpu() is set to true)"
            )
        print(f"Availabilities: {availablilities}. Picked Device {picked_device}")
        return picked_device

    @classmethod
    def _use_gpu(cls, index: int):
        size = cls.gpu_memory_limit()
        gpus = tf.config.experimental.list_physical_devices("GPU")

        # Check if selected gpu can be found
        if gpus is None or len(gpus) <= index:
            raise ValueError(
                f"Uncertainty Wizards DeviceAllocatorContextManager was configured to use gpu {index} "
                f"but no no such gpu was found.  "
            )

        try:
            tf.config.set_visible_devices([gpus[index]], "GPU")
            tf.config.experimental.set_virtual_device_configuration(
                gpus[index],
                [tf.config.experimental.VirtualDeviceConfiguration(memory_limit=size)],
            )
            logical_gpus = tf.config.experimental.list_logical_devices("GPU")
            print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
        except RuntimeError as e:
            raise ValueError(
                f"Uncertainty Wizard was unable to create a virtual device "
                f"on gpu {index} and memory limit {size}MB"
            ) from e

    @classmethod
    def _use_cpu(cls):
        try:
            # Disable all GPUS
            tf.config.set_visible_devices([], "GPU")
            visible_devices = tf.config.get_visible_devices()
            for device in visible_devices:
                assert device.device_type != "GPU"
        except RuntimeError as e:
            raise ValueError(
                f"Uncertainty Wizard was unable to disable gpu use."
            ) from e

    @classmethod
    def _acquire_lock(cls) -> int:
        """
        Waits until no lockfile is present and, once possible, creates a lockfile.
        Code inspired by https://github.com/dmfrey/FileLock/blob/master/filelock/filelock.py

        :returns the file descriptor (int) of the acquired lockfile
        :raise RuntimeError if the lock acquiring times out or if an IO error prevents lock acquiring.
        """
        timeout = cls.acquire_lock_timeout()
        start_time = time.time()
        while True:
            try:
                return os.open(
                    cls._lock_file_path(),
                    (
                        os.O_CREAT  # create file if it does not exist
                        | os.O_EXCL  # error if create and file exists
                        | os.O_RDWR
                    ),  # open for reading and writing
                )
            except OSError as e:
                if e.errno != errno.EEXIST:
                    # Some other error than 'file exists' occurred
                    raise RuntimeError(
                        "An error occurred when trying to acquire lock for device allocation "
                    ) from e
                if (time.time() - start_time) >= timeout:
                    raise RuntimeError(
                        (
                            f"Ensemble process was not capable of acquiring lock in {timeout} seconds."
                            f"Make sure that no file `{cls._lock_file_path()}` exists (delete it if it does)."
                            f"If this does not help, consider increasing the timeout by overriding `acquire_lock_timeout`"
                            f"in your DeviceAllocatorContextManager extension"
                        )
                    )
                time.sleep(cls.delay())

    @classmethod
    def _release_lock(cls, lockfile: int):
        os.close(lockfile)
        os.remove(cls._lock_file_path())



import abc
import enum
from typing import List, Union

import numpy as np


class ProblemType(enum.Enum):
    """
    An enum used to distinguish between regression and classification problems.
    Might be extended later, to support also other or more specific ProblemTypes
    (e.g. one-class classification)
    """

    REGRESSION = enum.auto()
    CLASSIFICATION = enum.auto()


class Quantifier(abc.ABC):
    """
    Quantifiers are dependencies, injectable into prediction calls,
    which calculate predictions and uncertainties or confidences
    from DNN outputs.

    The quantifier class is abstract and should not be directly implemented.
    Instead, new quantifiers should extend uwiz.quantifiers.ConfidenceQuantifier
    or uwiz.quantifiers.UncertaintyQuantifier instead.
    """

    @classmethod
    @abc.abstractmethod
    def aliases(cls) -> List[str]:
        """
        Aliases are string identifiers of this quantifier.
        They are used to select quantifiers by string in predict methods (need to be registered in quantifier_registry).

        Additionally, the first identifier in the list is used for logging purpose.
        Thus, the returned list have at least length 1.

        :return: list of quantifier identifiers
        """
        pass  # pragma: no cover

    @classmethod
    @abc.abstractmethod
    def is_confidence(cls) -> bool:
        """
        Boolean flag indicating whether this quantifier quantifies uncertainty or confidence.
        They are different as follows (assuming that the quantifier actually correctly captures the chance of misprediction):

        - In `uncertainty quantification`, the higher the quantification, the higher the chance of misprediction.
        - in `confidence quantification` the lower the quantification, the higher the change of misprediction.

         :return: True iff this is a confidence quantifier, False if this is an uncertainty quantifier

        """
        pass  # pragma: no cover

    @classmethod
    @abc.abstractmethod
    def takes_samples(cls) -> bool:
        """
        A flag indicating whether this quantifier relies on monte carlo samples
        (in which case the method returns True)
        or on a single neural network output
        (in which case the method return False)

        :return: True if this quantifier expects monte carlo samples for quantification. False otherwise.
        """
        pass  # pragma: no cover

    @classmethod
    @abc.abstractmethod
    def problem_type(cls) -> ProblemType:
        """
        Specifies whether this quantifier is applicable to classification or regression problems
        :return: One of the two enum values REGRESSION or CLASSIFICATION
        """
        pass  # pragma: no cover

    @classmethod
    @abc.abstractmethod
    def calculate(cls, nn_outputs: np.ndarray):
        """
        Calculates the predictions and uncertainties.


        Note this this assumes *batches* of neural network outputs.
        When using this method for a single nn output, make sure to reshape the passed array,
        e.g. using `x = np.expand_dims(x, axis=0)`

        The method returns a tuple of

        - A prediction (int or float) or array of predictions
        - A uncertainty or confidence quantification (float) or array of uncertainties

        :param nn_outputs: The NN outputs to be considered when determining prediction and uncertainty quantification
        :return: A tuple of prediction(s) and uncertainty(-ies).
        """
        pass  # pragma: no cover

    @classmethod
    def cast_conf_or_unc(
        cls, as_confidence: Union[None, bool], superv_scores: np.ndarray
    ) -> np.ndarray:
        """
        Utility method to convert confidence metrics into uncertainty and vice versa.
        Call `is_confidence()` to find out if this is a uncertainty or a confidence metric.

        The supervisors scores are converted as follows:

         - Confidences are multiplied by (-1) iff `as_confidence` is False
         - Uncertainties are multiplied by (-1) iff `as_confidence` is True
         - Otherwise, the passed supervisor scores are returned unchanged.


        :param as_confidence: : A boolean indicating if the scores should be converted to confidences (True) or uncertainties (False)
        :param superv_scores: : The scores that are to be converted, provided a conversion is needed.
        :return: The converted scores or the unchanged `superv_scores` (if `as_confidence` is None or no conversion is needed)

        """
        if as_confidence is not None and cls.is_confidence() != as_confidence:
            return superv_scores * -1
        return superv_scores


class ConfidenceQuantifier(Quantifier, abc.ABC):
    """
    An abstract Quantifier subclass, serving as superclass for all confidence quantifying quantifiers:
    In `confidence quantification` the lower the value, the higher the chance of misprediction.
    """

    # docstr-coverage:inherited
    @classmethod
    def is_confidence(cls) -> bool:
        return True


class UncertaintyQuantifier(Quantifier, abc.ABC):
    """
    An abstract Quantifier subclass, serving as superclass for all uncertainty quantifying quantifiers:
    In `uncertainty quantification` the lower the value, the lower the chance of misprediction.
    """

    # docstr-coverage:inherited
    @classmethod
    def is_confidence(cls) -> bool:
        return False
